﻿using System;
using System.ComponentModel;
using System.Text;
using System.Windows.Forms;
using ClipExtension.Common;

namespace ClipExtension.Controls
{
	public partial class DataUnitsViewer : UserControl
	{
		private const string NewLine = "\r\n";

		private MatchedItemPredicater MatchedItemPredicater;

		public DataUnitsViewer()
		{
			InitializeComponent();

			InitializeEtc();
		}

		[Category( "Layout" )]
		[Browsable( true )]
		public bool OptionsPanelCollapsed
		{
			get { return LeftSplitContainer.Panel1Collapsed; }
			set { LeftSplitContainer.Panel1Collapsed = value; }
		}

		[Category( "Layout" )]
		[Browsable( true )]
		public bool InformationPanelCollapsed
		{
			get { return SplitContainer1.Panel2Collapsed; }
			set { SplitContainer1.Panel2Collapsed = value; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public bool FirstMatching
		{
			get { return FirstMatchingCheckBox.Checked; }
			set { FirstMatchingCheckBox.Checked = value; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public bool CaseSensitive
		{
			get { return CaseSensitiveCheckBox.Checked; }
			set { CaseSensitiveCheckBox.Checked = value; }
		}

		[Category( "Behavior" )]
		[DefaultValue( SearchMode.Normal )]
		[Browsable( true )]
		public SearchMode SearchMode
		{
			get { return GetSearchMode(); }
			set { SetSearchMode( value ); }
		}

		public bool HasMigemo
		{
			get { return MatchedItemPredicater.HasMigemo; }
		}

		private void ComboBox_KeyDown( object sender, KeyEventArgs e )
		{
			if ( e.KeyData ==
			     ( Keys.Control | Keys.A ) )
			{
				e.SuppressKeyPress = true;
				TagsLabeledComboBox.ComboBox.SelectAll();
			}
		}

		private void TextBoxOrComboBoxOrListBoxKeyDown( object sender, KeyEventArgs e )
		{
			if ( e.KeyData ==
			     ( Keys.Control | Keys.Shift | Keys.A ) )
			{
				e.SuppressKeyPress = true;
				ContentsFilterListBox.SelectAllItems();
			}
		}

		private void SetSearchMode( SearchMode value )
		{
			switch ( value )
			{
				case SearchMode.Normal:
					NormalRadioButton.Checked = true;

					break;
				case SearchMode.Regex:
					RegexRadioButton.Checked = true;

					break;
				case SearchMode.Migemo:
					if ( MatchedItemPredicater.HasMigemo )
					{
						MigemoRadioButton.Checked = true;
					}
					else
					{
						NormalRadioButton.Checked = true;
					}

					break;
			}

			OnSearchOptionsChanged( EventArgs.Empty );
		}

		private SearchMode GetSearchMode()
		{
			if ( MigemoRadioButton.Checked )
			{
				return SearchMode.Migemo;
			}
			else if ( RegexRadioButton.Checked )
			{
				return SearchMode.Regex;
			}

			return SearchMode.Normal;
		}

		[Browsable( true )]
		public event EventHandler<EventArgs> SearchOptionsChanged;

		protected virtual void OnSearchOptionsChanged( EventArgs e )
		{
			if ( SearchOptionsChanged != null )
			{
				SearchOptionsChanged( this, e );
			}

			ContentsFilterListBox.UpdateMatchedItems();
		}

		private void InitializeEtc()
		{
			SearchMode = SearchMode.Normal;

			MatchedItemPredicater = new MatchedItemPredicater();

			ContentsFilterListBox.SelectedIndexChanged += ContentsFilterListBox_SelectedIndexChanged;
			ContentsFilterListBox.MatchedItemsUpdating += ContentsFilterListBox_MatchedItemsUpdating;
			ContentsFilterListBox.SetMatchedItemPredicater( MatchedItemPredicater );
			ContentsFilterListBox.SetMatchedItemComparer( new ItemComparer() );

			ContentsFilterListBox.ListBoxKeyDown += TextBoxOrComboBoxOrListBoxKeyDown;
			ContentsFilterListBox.TextBoxKeyDown += TextBoxOrComboBoxOrListBoxKeyDown;
			TagsLabeledComboBox.ComboBox.KeyDown += TextBoxOrComboBoxOrListBoxKeyDown;

			TagsLabeledComboBox.ComboBox.TextChanged += TagsChanged;
			TagsLabeledComboBox.ComboBox.KeyDown += ComboBox_KeyDown;

			if ( !MatchedItemPredicater.HasMigemo )
			{
				MigemoRadioButton.Enabled = false;
			}
		}

		private void ContentsFilterListBox_MatchedItemsUpdating( object sender, EventArgs e )
		{
			MatchedItemPredicater.PrepareIsMatchedItemArgs(
				ContentsFilterListBox.FilterText,
				TagsLabeledComboBox.ComboBox.Text,
				SearchMode,
				FirstMatching,
				CaseSensitive
				);
		}

		private void ContentsFilterListBox_SelectedIndexChanged( object sender, EventArgs e )
		{
			UpdateInformation();
		}

		private void UpdateInformation()
		{
			var selectedItem = (DataUnit) ContentsFilterListBox.GetOnlyOneSelectedItem();

			if ( selectedItem == null )
			{
				InformationTextBox.Text = string.Empty;
			}
			else
			{
				InformationTextBox.Text = GetInformation( selectedItem );
			}
		}

		private string GetInformation( DataUnit selectedItem )
		{
			var result = new StringBuilder();

			result.Append( selectedItem.ToString() );

			if ( selectedItem.Tags.Count == 0 )
			{
				return result.ToString();
			}

			result.Append( NewLine );
			result.Append( NewLine );

			result.Append( "タグ:" );
			result.Append( NewLine );
			result.Append( string.Join( NewLine, selectedItem.Tags ) );

			return result.ToString();
		}

		private void NormalAndRegexRadioButton_CheckedChanged( object sender, EventArgs e )
		{
			OnSearchOptionsChanged( EventArgs.Empty );
		}

		private void MigemoRadioButton_CheckedChanged( object sender, EventArgs e )
		{
			CaseSensitiveCheckBox.Enabled = !MigemoRadioButton.Checked;
			OnSearchOptionsChanged( EventArgs.Empty );
		}

		private void CheckBox_CheckedChanged( object sender, EventArgs e )
		{
			OnSearchOptionsChanged( EventArgs.Empty );
		}

		private void TagsChanged( object sender, EventArgs e )
		{
			ContentsFilterListBox.UpdateMatchedItems();
		}

		public override string ToString()
		{
			var result = new StringBuilder();

			switch ( SearchMode )
			{
				case SearchMode.Normal:
					result.Append( "通常" );
					break;
				case SearchMode.Regex:
					result.Append( "正規表現" );
					break;
				case SearchMode.Migemo:
					result.Append( "Migemo" );
					break;
				default:
					throw new ArgumentOutOfRangeException();
			}

			if ( FirstMatching )
			{
				result.Append( ", 先頭一致" );
			}

			if ( CaseSensitive && SearchMode != SearchMode.Migemo )
			{
				result.Append( ", 大文字・小文字の区別" );
			}

			return result.ToString();
		}
	}
}