﻿using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using ClipExtension.Common;
using KaoriYa.Migemo;
using Tomham;
using Tomham.Windows.Forms;

namespace ClipExtension.Controls
{
	public class MatchedItemPredicater : IMatchedItemPredicater
	{
		#region Migemo

		private static readonly Migemo Migemo = CreateMigemo();

		private static Migemo CreateMigemo()
		{
			const string path = @"dict\migemo-dict";
			try
			{
				return new Migemo( path );
			}
			catch ( Exception e )
			{
				Console.WriteLine( e.Message );

				return null;
			}
		}

		#endregion

		public bool HasMigemo
		{
			get { return Migemo != null; }
		}

		#region IsMatchedItem

		private const int MaxPatternLength = 500;
		private IsMatchedItemArgs Args;

		public bool IsMatchedItem( string filterText, object item )
		{
			var dataUnit = (DataUnit) item;
			return IsMatchedItem( dataUnit );
		}

		public void PrepareIsMatchedItemArgs( string filter,
		                                      string tags,
		                                      SearchMode searchMode,
		                                      bool initialSearch,
		                                      bool caseSensitive )
		{
			Args = new IsMatchedItemArgs
			       { SearchMode = searchMode, FirstMatching = initialSearch, CaseSensitive = caseSensitive };

			Args.Tags = Split( tags );

			switch ( searchMode )
			{
				case SearchMode.Normal:
					Args.Words = Split( filter );

					break;

				case SearchMode.Regex:
					Args.Words.Add( filter );
					PrepareRegexes();
					break;

				case SearchMode.Migemo:
					Args.Words = Split( filter );
					PrepareMigemo();
					break;

				default:
					throw new ArgumentOutOfRangeException( "searchMode" );
			}
		}

		private void PrepareRegexes()
		{
			string word = Args.Words[0];
			CreateAndAddRegex( word, word );
		}

		private void PrepareMigemo()
		{
			if ( Args.SearchMode == SearchMode.Migemo &&
			     HasMigemo )
			{
				foreach ( string word in Args.Words )
				{
					string pattern = Migemo.Query( word );

					CreateAndAddRegex( word, pattern );
				}
			}
		}

		private void CreateAndAddRegex( string word, string pattern )
		{
			if ( Args.FirstMatching )
			{
				pattern = @"^" + pattern;
			}

			if ( pattern.Length < MaxPatternLength )
			{
				try
				{
					if ( Args.CaseSensitive )
					{
						Args.Regexes.Add( word, new Regex( pattern ) );
					}
					else
					{
						Args.Regexes.Add( word, new Regex( pattern, RegexOptions.IgnoreCase ) );
					}
				}
				catch ( ArgumentException e )
				{
					Console.WriteLine( e.Message );
				}
			}
		}

		private bool IsMatchedItem( DataUnit dataUnit )
		{
			if ( Args.Words.Count == 0 &&
			     Args.Tags.Count == 0 )
			{
				return true;
			}

			if ( !HasAllTags( dataUnit ) )
			{
				return false;
			}

			if ( !IsMatchedItemWithAllWords( dataUnit ) )
			{
				return false;
			}

			return true;
		}

		private bool IsMatchedItemWithAllWords( DataUnit dataUnit )
		{
			foreach ( string word in Args.Words )
			{
				if ( !IsMatchedItemWithWord( dataUnit, word ) )
				{
					return false;
				}
			}

			return true;
		}

		private bool IsMatchedItemWithWord( DataUnit dataUnit, string word )
		{
			switch ( Args.SearchMode )
			{
				case SearchMode.Normal:
					return IsMatchedItemNormal( dataUnit, word );

				case SearchMode.Migemo:
					return IsMatchedItemRegex( dataUnit, word );

				case SearchMode.Regex:
					return IsMatchedItemRegex( dataUnit, word );

				default:
					return false;
			}
		}

		private bool IsMatchedItemNormal( DataUnit dataUnit, string word )
		{
			if ( Args.FirstMatching )
			{
				if ( Args.CaseSensitive )
				{
					return dataUnit.ToString().StartsWith( word );
				}
				else
				{
					return dataUnit.ToString().StartsWith( word, StringComparison.CurrentCultureIgnoreCase );
				}
			}

			if ( Args.CaseSensitive )
			{
				return dataUnit.ToString().Contains( word );
			}
			else
			{
				return dataUnit.ToString().ToLower().Contains( word.ToLower() );
			}
		}

		private bool IsMatchedItemRegex( DataUnit dataUnit, string word )
		{
			if ( !Args.Regexes.ContainsKey( word ) )
			{
				return false;
			}

			Regex regex = Args.Regexes[word];

			try
			{
				return regex.IsMatch( dataUnit.ToString() );
			}
			catch ( Exception e )
			{
				Console.WriteLine( e.Message );
				return false;
			}
		}

		#region HasAllTags

		private bool HasAllTags( DataUnit dataUnit )
		{
			foreach ( string tag in Args.Tags )
			{
				if ( !HasTag( dataUnit, tag ) )
				{
					return false;
				}
			}

			return true;
		}

		private bool HasTag( DataUnit dataUnit, string tag )
		{
			foreach ( string one in dataUnit.Tags )
			{
				if ( one.StartsWith( tag, StringComparison.CurrentCultureIgnoreCase ) )
				{
					return true;
				}
			}

			return false;
		}

		#endregion

		#endregion

		#region Split

		private List<string> Split( string text )
		{
			string trimmed = text.Trim();

			var result = new List<string>();

			while ( 0 < trimmed.Length )
			{
				string element = CutOffFirstElement( ref trimmed );

				if ( element != string.Empty )
				{
					result.Add( element );
				}
			}

			return result;
		}

		private string CutOffFirstElement( ref string text )
		{
			string result;

			if ( text[0] ==
			     CharConstants.DOUBLE_QUOTATION )
			{
				text = text.Remove( 0, 1 );

				result = DoCutOffFirstElement( ref text, CharConstants.DOUBLE_QUOTATION );
			}
			else
			{
				result = DoCutOffFirstElement( ref text, CharConstants.SPACE );
			}

			text = text.Trim();

			return result;
		}

		private string DoCutOffFirstElement( ref string text, char endOfElement )
		{
			string result;

			int endOfElementIndex = text.IndexOf( endOfElement );

			if ( endOfElementIndex == -1 )
			{
				result = text.Trim();

				text = string.Empty;
			}
			else
			{
				result = text.Substring( 0, endOfElementIndex );

				text = text.Remove( 0, endOfElementIndex + 1 );
			}

			return result;
		}

		#endregion
	}
}