﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.ComponentModel.Composition.Hosting;
using System.Threading;
using System.Windows.Forms;
using ClipExtension.Common;
using ClipExtension.Plugins;
using Tomham;
using Tomham.Windows.Forms;

namespace ClipExtension.Main
{
	public class ClipboardManipulator : IDisposable, IClipboardManipulator
	{
		private readonly DirectoryHelper DirectoryHelper;
		private readonly IMessageOutput Output;
		private readonly RepositorySerializer RepositorySerializer = new RepositorySerializer();
		private int BeforePasteDelayStore;
		private int CopyReTryDelayStore;
		private int CopyReTryTimesStore;

		public ClipboardManipulator( IMessageOutput output,
		                             CompositionContainer container,
		                             DirectoryHelper directoryHelper,
		                             PluginsPreference processorsPluginsPreference,
		                             PluginsPreference formatSupportersPluginsPreference,
		                             List<Paster> pasters,
		                             ICommandTable<ClipboardManipulateCommandArgs> clipboardManipulateCommandTable,
		                             CommandTable<EditCommandArgs> editCommandTable,
		                             ICommandTable<GeneralCommandArgs> generalCommandTable )
		{
			Output = output;

			DirectoryHelper = directoryHelper;

			ClipboardWatcher = new ClipboardWatcher();

			container.ComposeParts( this );

			InitializeFormatSupporters( formatSupportersPluginsPreference );
			InitializeProcessors( processorsPluginsPreference );
			LoadRepository();
			LoadPlugins( generalCommandTable, clipboardManipulateCommandTable, editCommandTable );

			PasterUnit = new PasterUnit { Pasters = pasters };

			ClipboardWatcher.DrawClipboard += ClipboardWatcherDrawClipboard;
		}

		public bool Enabled
		{
			get { return ClipboardWatcher.Enabled; }
			set { ClipboardWatcher.Enabled = value; }
		}

		public FormatSupporterUnit FormatSupporterUnit { get; private set; }

		[ImportMany]
		public List<Lazy<IFormatSupporter, IPluginMetadata>> InitialFormatSupporters { get; set; }

		[ImportMany]
		public List<Lazy<IProcessor, IPluginMetadata>> InitialProcessors { get; set; }

		public PluginContainer<IProcessor> Processors { get; private set; }

		private ClipboardWatcher ClipboardWatcher { get; set; }

		public bool CheckAndRegisterChainEnabled
		{
			get { return ClipboardWatcher.CheckAndRegisterChainEnabled; }
			set { ClipboardWatcher.CheckAndRegisterChainEnabled = value; }
		}

		public int CheckAndRegisterChainInterval
		{
			get { return ClipboardWatcher.CheckAndRegisterChainInterval; }
			set { ClipboardWatcher.CheckAndRegisterChainInterval = value; }
		}

		public PasterUnit PasterUnit { get; private set; }

		#region IClipboardManipulator Members

		public int BeforePasteDelay
		{
			get
			{
				if ( BeforePasteDelayStore < 50 )
				{
					BeforePasteDelayStore = 50;
				}
				else if ( 1000 < BeforePasteDelayStore )
				{
					BeforePasteDelayStore = 1000;
				}

				return BeforePasteDelayStore;
			}

			set { BeforePasteDelayStore = value; }
		}

		public int CopyReTryDelay
		{
			get
			{
				if ( CopyReTryDelayStore < 50 )
				{
					CopyReTryDelayStore = 50;
				}
				else if ( 1000 < CopyReTryDelayStore )
				{
					CopyReTryDelayStore = 1000;
				}
				return CopyReTryDelayStore;
			}

			set { CopyReTryDelayStore = value; }
		}

		public int CopyReTryTimes
		{
			get
			{
				if ( CopyReTryTimesStore < 0 )
				{
					CopyReTryTimesStore = 0;
				}
				else if ( 20 < CopyReTryTimesStore )
				{
					CopyReTryTimesStore = 20;
				}
				return CopyReTryTimesStore;
			}

			set { CopyReTryTimesStore = value; }
		}

		public Repository Repository { get; set; }

		public void CopyAndPaste( DataUnit dataUnit )
		{
			if ( dataUnit == null )
			{
				throw new ArgumentNullException( "dataUnit" );
			}

			if ( Copy( dataUnit ) )
			{
				Paste();
			}
		}

		public bool Copy( DataUnit dataUnit )
		{
			if ( dataUnit == null )
			{
				throw new ArgumentNullException( "dataUnit" );
			}

			dataUnit.UpdateLastUsedTime();

			DataUnit clone = dataUnit.Clone();
			try
			{
				if ( !CopyingWithProcessors( clone ) )
				{
					return false;
				}

				object dataObject = FormatSupporterUnit.CreateDataObject( clone );
				Clipboard.SetDataObject( dataObject, true, CopyReTryTimes, CopyReTryDelay );

				CopiedWithProcessors( clone );

				return true;
			}
			catch ( Exception e )
			{
				WriteExceptionInformaition( e );
			}

			return false;
		}

		#endregion

		#region IDisposable Members

		public void Dispose()
		{
			ClipboardWatcher.Dispose();

			SaveRepository();
			UnloadPlugins();
		}

		#endregion

		#region SaveRepository

		public void SaveRepository()
		{
			Repository.Adjust();

			RepositorySavingPlugins();

			RepositorySerializer.Serialize( Repository, DirectoryHelper );

			RepositorySavedPlugins();
		}

		private void RepositorySavingPlugins()
		{
			var args = new RepositorySavingArgs( Output, Repository );
			foreach ( var formatSupporter in FormatSupporterUnit.PluginContainer.Items )
			{
				try
				{
					formatSupporter.Value.RepositorySaving( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}
			}

			foreach ( var processor in Processors.Items )
			{
				try
				{
					processor.Value.RepositorySaving( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}
			}
		}

		private void RepositorySavedPlugins()
		{
			var args = new RepositorySavedArgs( Output, Repository );
			foreach ( var formatSupporter in FormatSupporterUnit.PluginContainer.Items )
			{
				try
				{
					formatSupporter.Value.RepositorySaved( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}
			}

			foreach ( var processor in Processors.Items )
			{
				try
				{
					processor.Value.RepositorySaved( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}
			}
		}

		#endregion

		private void InitializeFormatSupporters( PluginsPreference pluginsPreference )
		{
			FormatSupporterUnit = new FormatSupporterUnit( InitialFormatSupporters, pluginsPreference, Output );
		}

		private void InitializeProcessors( PluginsPreference pluginsPreference )
		{
			Processors = new PluginContainer<IProcessor>( InitialProcessors, pluginsPreference );
		}

		private void LoadRepository()
		{
			Repository = RepositorySerializer.DeSerialize( DirectoryHelper );
		}

		private void ClipboardWatcherDrawClipboard( object sender, DrawClipboardEventArgs e )
		{
			CopyFromClipboard();
		}

		public void CopyFromClipboard()
		{
			try
			{
				IDataObject dataObject = Clipboard.GetDataObject();

				if ( dataObject == null )
				{
					return;
				}

				DataUnit dataUnit = FormatSupporterUnit.CreateDataUnit( dataObject );
				if ( dataUnit == null )
				{
					return;
				}

				if ( !CopyingFromClipboardWithProcessors( dataUnit ) )
				{
					return;
				}

				Repository.PushToHistory( dataUnit );

				CopiedFromClipboardWithProcessors( dataUnit );
			}
			catch ( Exception exeception )
			{
				WriteExceptionInformaition( exeception );
			}
		}

		private void Paste()
		{
			Thread.Sleep( BeforePasteDelay );

			PasterUnit.Paste();
		}

		private void WriteExceptionInformaition( Exception e )
		{
			Output.WriteLine( e.Message );
			Output.WriteLine( e.StackTrace );
		}

		public void CheckAndRegisterChain()
		{
			ClipboardWatcher.CheckAndRegisterChain();
		}

		#region LoadPlugins

		private void LoadPlugins( ICommandTable<GeneralCommandArgs> generalCommandTable,
		                          ICommandTable<ClipboardManipulateCommandArgs> clipboardManipulateCommandTable,
		                          CommandTable<EditCommandArgs> editCommandTable )
		{
			LoadProcessors( generalCommandTable, clipboardManipulateCommandTable, editCommandTable );
			LoadFormatSupporters( generalCommandTable, clipboardManipulateCommandTable, editCommandTable );
		}

		private void LoadFormatSupporters( ICommandTable<GeneralCommandArgs> generalCommandTable,
		                                   ICommandTable<ClipboardManipulateCommandArgs> clipboardManipulateCommandTable,
		                                   CommandTable<EditCommandArgs> editCommandTable )
		{
			var args = new LoadArgs( Output, Repository, generalCommandTable, clipboardManipulateCommandTable, editCommandTable );
			foreach ( var formatSupporter in FormatSupporterUnit.PluginContainer.Items )
			{
				formatSupporter.Value.Load( args );
			}
		}

		private void LoadProcessors( ICommandTable<GeneralCommandArgs> generalCommandTable,
		                             ICommandTable<ClipboardManipulateCommandArgs> clipboardManipulateCommandTable,
		                             CommandTable<EditCommandArgs> editCommandTable )
		{
			var args = new LoadArgs( Output, Repository, generalCommandTable, clipboardManipulateCommandTable, editCommandTable );
			foreach ( var processor in Processors.Items )
			{
				processor.Value.Load( args );
			}
		}

		#endregion

		#region UnloadPlugins

		private void UnloadPlugins()
		{
			UnLoadProcessors();
			UnLoadFormatSupporters();
		}

		private void UnLoadFormatSupporters()
		{
			var args = new UnLoadArgs();
			foreach ( var formatSupporter in FormatSupporterUnit.PluginContainer.Items )
			{
				formatSupporter.Value.UnLoad( args );
			}
		}

		private void UnLoadProcessors()
		{
			var args = new UnLoadArgs();
			foreach ( var processor in Processors.Items )
			{
				processor.Value.UnLoad( args );
			}
		}

		#endregion

		#region Action with Processors

		/// <summary>
		/// 
		/// </summary>
		/// <param name="dataUnit"></param>
		/// <returns>処理を続行して良いか</returns>
		private bool CopyingFromClipboardWithProcessors( DataUnit dataUnit )
		{
			var args = new CopyingFromClipboardArgs( Output, Repository, dataUnit );
			foreach ( var processor in Processors.Items )
			{
				try
				{
					processor.Value.CopyingFromClipboard( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}

				if ( args.Cancel )
				{
					return false;
				}
			}

			return true;
		}

		private void CopiedFromClipboardWithProcessors( DataUnit dataUnit )
		{
			var args = new CopiedFromClipboardArgs( Output, Repository, dataUnit );
			foreach ( var processor in Processors.Items )
			{
				try
				{
					processor.Value.CopiedFromClipboard( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}
			}
		}

		private bool CopyingWithProcessors( DataUnit dataUnit )
		{
			var args = new CopyingArgs( Output, Repository, dataUnit );
			foreach ( var processor in Processors.Items )
			{
				try
				{
					processor.Value.Copying( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}

				if ( args.Cancel )
				{
					return false;
				}
			}

			return true;
		}

		private void CopiedWithProcessors( DataUnit dataUnit )
		{
			var args = new CopiedArgs( Output, Repository, dataUnit );
			foreach ( var processor in Processors.Items )
			{
				try
				{
					processor.Value.Copied( args );
				}
				catch ( Exception e )
				{
					WriteExceptionInformaition( e );
				}
			}
		}

		#endregion
	}
}