﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using ClipExtension.Common;
using Tomham.IO;
using XmlDeSerializer = Yaowi.Common.Serialization.XmlDeserializer;

namespace ClipExtension.Main
{
	public class RepositorySerializer
	{
		private readonly List<Tuple<DataUnit, string>> FavoritesDataUnitToFileNameTable = new List<Tuple<DataUnit, string>>();
		private readonly List<Tuple<DataUnit, string>> HistoryDataUnitToFileNameTable = new List<Tuple<DataUnit, string>>();

		#region Convert

		private DataUnitSerializationItem Convert( DataUnit dataUnit )
		{
			var item = new DataUnitSerializationItem { Target = dataUnit };
			item.UpdateFromTarget();

			return item;
		}

		private DataUnit Convert( DataUnitSerializationItem item )
		{
			item.UpdateToTarget();
			return item.Target;
		}

		#endregion

		public void Serialize( Repository repository, DirectoryHelper directoryHelper )
		{
			var serializer = new DataUnitSerializationItemSerializer( SerializerHelper.GetXmlSerializer(),
			                                                          SerializerHelper.GetXmlDeserializer(
			                                                          	directoryHelper.PluginsDirectory ) );

			Serialize( repository.History, directoryHelper.HistoryDirectory, HistoryDataUnitToFileNameTable, serializer );
			Serialize( repository.Favorites, directoryHelper.FavoritesDirectory, FavoritesDataUnitToFileNameTable, serializer );
		}

		private void Serialize( IEnumerable<DataUnit> data,
		                        string directory,
		                        List<Tuple<DataUnit, string>> table,
		                        DataUnitSerializationItemSerializer serializer )
		{
			UpdateTable( data, directory, table );

			RemoveNeedLessFiles( table, directory );

			DoSerialize( table, serializer );
		}

		private void DoSerialize( IEnumerable<Tuple<DataUnit, string>> table, DataUnitSerializationItemSerializer serializer )
		{
			foreach ( var tuple in table )
			{
				DataUnitSerializationItem item = Convert( tuple.Item1 );

				serializer.Serialize( item, tuple.Item2 );
			}
		}

		private void UpdateTable( IEnumerable<DataUnit> data,
		                          string directory,
		                          List<Tuple<DataUnit, string>> table )
		{
			var newTable = new List<Tuple<DataUnit, string>>();

			int number = 0;

			foreach ( DataUnit dataUnit in data )
			{
				string fileName = GetFileName( dataUnit, directory, table, ref number );

				newTable.Add( new Tuple<DataUnit, string>( dataUnit, fileName ) );
			}

			table.Clear();
			table.AddRange( newTable );
		}

		private string GetFileName( DataUnit dataUnit,
		                            string directory,
		                            IEnumerable<Tuple<DataUnit, string>> table,
		                            ref int number )
		{
			foreach ( var tuple in table )
			{
				if ( ReferenceEquals( tuple.Item1, dataUnit ) )
				{
					return tuple.Item2;
				}
			}

			string result = DateTime.Now.ToString( "yyyyMMddHHmmss" );
			result += string.Format( "-{0:D10}", number ) + ".dat";

			result = Path.Combine( directory, result );
			number++;

			return result;
		}

		private void RemoveNeedLessFiles( List<Tuple<DataUnit, string>> table, string directory )
		{
			string[] files = Directory.GetFiles( directory, "*.dat" );

			for ( int i = 0; i < files.Length; i++ )
			{
				string fileName = files[i];

				if ( !Contains( table, fileName ) )
				{
					FileHelper.ForceDeleteFile( fileName );
				}
			}
		}

		private bool Contains( IEnumerable<Tuple<DataUnit, string>> table, string fileName )
		{
			return table.Any( tuple => tuple.Item2 == fileName );
		}

		#region DeSerialize

		public Repository DeSerialize( DirectoryHelper directoryHelper )
		{
			var result = new Repository();
			var serializer = new DataUnitSerializationItemSerializer( SerializerHelper.GetXmlSerializer(),
			                                                          SerializerHelper.GetXmlDeserializer(
			                                                          	directoryHelper.PluginsDirectory ) );

			result.History = DeSerialize( directoryHelper.HistoryDirectory, HistoryDataUnitToFileNameTable, serializer );
			result.Favorites = DeSerialize( directoryHelper.FavoritesDirectory, FavoritesDataUnitToFileNameTable, serializer );

			return result;
		}

		private List<DataUnit> DeSerialize( string directory,
		                                    List<Tuple<DataUnit, string>> table,
		                                    DataUnitSerializationItemSerializer serializer )
		{
			table.Clear();

			var result = new List<DataUnit>();

			string[] files = Directory.GetFiles( directory, "*.dat" );

			foreach ( string fileName in files )
			{
				try
				{
					DataUnitSerializationItem item = serializer.DeSerialize( fileName );

					if ( item == null )
					{
						continue;
					}

					DataUnit dataUnit = Convert( item );
					result.Add( dataUnit );
					table.Add( new Tuple<DataUnit, string>( dataUnit, fileName ) );
				}
				catch ( Exception e )
				{
					Console.WriteLine( e.Message );
					Console.WriteLine( e.StackTrace );
				}
			}

			return result;
		}

		#endregion
	}
}