﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using ClipExtension.Common;
using ClipExtension.Controls;
using ClipExtension.Main;
using ClipExtension.Options;
using ClipExtension.Plugins;
using Tomham;
using Tomham.Windows.Forms;

namespace ClipExtension
{
	public partial class IncrementalSearchForm : Form
	{
		private readonly CommandTable<ClipboardManipulateCommandArgs> CommandTable;

		private readonly IncrementalSearchFormOptions Options;

		private DataUnitsViewer DataUnitsViewer;

		public IncrementalSearchForm( IMessageOutput output,
		                              ClipboardManipulator clipboardManipulator,
		                              IncrementalSearchFormOptions options,
		                              IEnumerable<DataUnit> source,
		                              CommandTable<ClipboardManipulateCommandArgs> commandTable,
		                              List<string> tags )
		{
			InitializeControls();

			Options = options;
			CommandTable = commandTable;

			Output = output;
			ClipboardManipulator = clipboardManipulator;

			DataUnitsViewer.ContentsFilterListBox.SetSource( source );
			DataUnitsViewer.ContentsFilterListBox.ListBoxDoubleClick += ContentsFilterListBox_ListBoxDoubleClick;
			DataUnitsViewer.TagsLabeledComboBox.ComboBox.Items.AddRange( tags.ToArray() );
			DataUnitsViewer.OptionsPanelCollapsed = true;
			if ( !DataUnitsViewer.HasMigemo )
			{
				MigemoToolStripMenuItem.Enabled = false;
			}

			UpdateFromOptions();

			DataUnitsViewer.SearchOptionsChanged += DataUnitsViewerSearchOptionsChanged;
		}

		private IMessageOutput Output { get; set; }

		private ClipboardManipulator ClipboardManipulator { get; set; }

		private void DataUnitsViewerSearchOptionsChanged( object sender, EventArgs e )
		{
			UpdateTextAndToolStripItems();
		}

		private void ContentsFilterListBox_ListBoxDoubleClick( object sender, EventArgs e )
		{
			CopyAndPaste();
		}

		private void UpdateToOptitons()
		{
			Options.MaxShowItemCount = DataUnitsViewer.ContentsFilterListBox.MaxShowItemCount;

			Options.InformationPanelCollapsed = DataUnitsViewer.InformationPanelCollapsed;

			Options.SearchMode = DataUnitsViewer.SearchMode;
			Options.FirstMatching = DataUnitsViewer.FirstMatching;
			Options.CaseSensitive = DataUnitsViewer.CaseSensitive;
		}

		private void UpdateFromOptions()
		{
			DataUnitsViewer.ContentsFilterListBox.TextBackColor = Options.BackColor;
			DataUnitsViewer.ContentsFilterListBox.TextForeColor = Options.ForeColor;
			DataUnitsViewer.ContentsFilterListBox.Font = (Font) Options.Font.Clone();

			DataUnitsViewer.TagsLabeledComboBox.ComboBox.BackColor = Options.BackColor;
			DataUnitsViewer.TagsLabeledComboBox.ComboBox.ForeColor = Options.ForeColor;
			DataUnitsViewer.TagsLabeledComboBox.ComboBox.Font = (Font) Options.Font.Clone();

			DataUnitsViewer.InformationTextBox.BackColor = Options.BackColor;
			DataUnitsViewer.InformationTextBox.ForeColor = Options.ForeColor;
			DataUnitsViewer.InformationTextBox.Font = (Font) Options.Font.Clone();

			DataUnitsViewer.ContentsFilterListBox.MaxShowItemCount = Options.MaxShowItemCount;

			DataUnitsViewer.InformationPanelCollapsed = Options.InformationPanelCollapsed;

			DataUnitsViewer.SearchMode = Options.SearchMode;
			DataUnitsViewer.FirstMatching = Options.FirstMatching;
			DataUnitsViewer.CaseSensitive = Options.CaseSensitive;

			UpdateTextAndToolStripItems();
		}

		private void InitializeControls()
		{
			InitializeComponent();

			DataUnitsViewer = new DataUnitsViewer { Dock = DockStyle.Fill };
			Controls.Add( DataUnitsViewer );
			Controls.SetChildIndex( DataUnitsViewer, 0 );

			DataUnitsViewer.ContentsFilterListBox.ListBoxKeyDown += ComboBoxOrListBoxKeyDown;
			DataUnitsViewer.ContentsFilterListBox.TextBoxKeyDown += ComboBoxOrListBoxKeyDown;
			DataUnitsViewer.ContentsFilterListBox.SelectionMode = SelectionMode.MultiExtended;
			DataUnitsViewer.TagsLabeledComboBox.ComboBox.KeyDown += ComboBoxOrListBoxKeyDown;
		}

		private void ComboBoxOrListBoxKeyDown( object sender, KeyEventArgs e )
		{
			if ( e.KeyData ==
			     Options.CopyAndPasteKey )
			{
				e.SuppressKeyPress = true;
				CopyAndPaste();
			}
			else if ( e.KeyData ==
			          Options.CopyKey )
			{
				e.SuppressKeyPress = true;
				Copy();
			}
			else if ( e.KeyData ==
			          Options.ShowCommandSearchFormKey )
			{
				e.SuppressKeyPress = true;
				ShowCommandSearchForm();
			}
		}

		private void ShowCommandSearchForm()
		{
			if ( !DataUnitsViewer.ContentsFilterListBox.HasSelectedItems() )
			{
				return;
			}

			HideAndMore();
			List<DataUnit> targets = DataUnitsViewer.ContentsFilterListBox.GetSelectedItems<DataUnit>();

			using ( var commandSearchForm = new SelectForm { Text = "コマンド" } )
			{
				commandSearchForm.SetSource( CommandTable.Commands );

				if ( commandSearchForm.ShowDialog() ==
				     DialogResult.OK )
				{
					Thread.Sleep( 50 );

					var args = new ClipboardManipulateCommandArgs( Output,
					                                               ClipboardManipulator,
					                                               targets );

					var selected = (Command<ClipboardManipulateCommandArgs>) commandSearchForm.SelectedItems[0];

					try
					{
						selected.Execute( args );
					}
					catch ( Exception e )
					{
						MessageBox.Show( e.Message );
						Output.WriteLine( e.Message );
						Output.WriteLine( e.StackTrace );
					}
				}
			}
		}

		private void HideAndMore()
		{
			Hide();
			UpdateToOptitons();
		}

		private void Copy()
		{
			var dataUnit = (DataUnit) DataUnitsViewer.ContentsFilterListBox.GetOnlyOneSelectedItem();

			if ( dataUnit != null )
			{
				HideAndMore();
				ClipboardManipulator.Copy( dataUnit );
			}
		}

		private void CopyAndPaste()
		{
			var dataUnit = (DataUnit) DataUnitsViewer.ContentsFilterListBox.GetOnlyOneSelectedItem();

			if ( dataUnit != null )
			{
				HideAndMore();
				ClipboardManipulator.CopyAndPaste( dataUnit );
			}
		}

		private void IncrementalSearchForm_Shown( object sender, EventArgs e )
		{
			DataUnitsViewer.ContentsFilterListBox.FocusToTextBox();
		}

		private void IncrementalSearchForm_KeyDown( object sender, KeyEventArgs e )
		{
			if ( e.KeyData.Equals( Keys.F | Keys.Control ) )
			{
				e.SuppressKeyPress = true;
				DataUnitsViewer.ContentsFilterListBox.FocusToTextBox();
			}

			else if ( e.KeyData.Equals( Keys.T | Keys.Control ) )
			{
				e.SuppressKeyPress = true;
				DataUnitsViewer.TagsLabeledComboBox.ComboBox.Focus();
			}
			else if ( e.KeyData.Equals( Keys.L | Keys.Control ) )
			{
				e.SuppressKeyPress = true;
				DataUnitsViewer.ContentsFilterListBox.FocusToListBox();
			}
		}

		private void IncrementalSearchForm_FormClosed( object sender, FormClosedEventArgs e )
		{
			UpdateToOptitons();
		}

		private void UpdateTextAndToolStripItems()
		{
			Text = DataUnitsViewer.ToString();

			NormalToolStripMenuItem.Checked = false;
			RegexToolStripMenuItem.Checked = false;
			MigemoToolStripMenuItem.Checked = false;

			switch ( DataUnitsViewer.SearchMode )
			{
				case SearchMode.Normal:
					NormalToolStripMenuItem.Checked = true;
					break;

				case SearchMode.Regex:
					RegexToolStripMenuItem.Checked = true;
					break;

				case SearchMode.Migemo:
					MigemoToolStripMenuItem.Checked = true;
					break;
			}

			CaseSensitiveToolStripMenuItem.Enabled = !DataUnitsViewer.SearchMode.Equals( SearchMode.Migemo );

			FirstMatchingToolStripMenuItem.Checked = DataUnitsViewer.FirstMatching;
			CaseSensitiveToolStripMenuItem.Checked = DataUnitsViewer.CaseSensitive;
		}

		#region ToolStripMenuItem Click events

		private void ToggleInformationPanelToolStripMenuItem_Click( object sender, EventArgs e )
		{
			DataUnitsViewer.InformationPanelCollapsed = !DataUnitsViewer.InformationPanelCollapsed;
		}

		private void NormalToolStripMenuItem_Click( object sender, EventArgs e )
		{
			DataUnitsViewer.SearchMode = SearchMode.Normal;
		}

		private void RegexToolStripMenuItem_Click( object sender, EventArgs e )
		{
			DataUnitsViewer.SearchMode = SearchMode.Regex;
		}

		private void MigemoToolStripMenuItem_Click( object sender, EventArgs e )
		{
			DataUnitsViewer.SearchMode = SearchMode.Migemo;
		}

		private void FirstMatchingToolStripMenuItem_Click( object sender, EventArgs e )
		{
			DataUnitsViewer.FirstMatching = !DataUnitsViewer.FirstMatching;
		}

		private void CaseSensitiveToolStripMenuItem_Click( object sender, EventArgs e )
		{
			DataUnitsViewer.CaseSensitive = !DataUnitsViewer.CaseSensitive;
		}

		#endregion
	}
}