﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.ComponentModel.Composition.Hosting;
using System.Drawing;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using ClipExtension.Common;
using ClipExtension.Main;
using ClipExtension.Options;
using ClipExtension.Plugins;
using Microsoft.Win32;
using Tomham;
using Tomham.Collections.Specialized;
using Tomham.IO;
using Tomham.Windows.Forms;
using Tomham.Windows.Forms.AutoGeneration;

namespace ClipExtension
{
	public partial class MainForm : Form
	{
		private readonly CommandTable<ClipboardManipulateCommandArgs> ClipboardManipulateCommandTable =
			new CommandTable<ClipboardManipulateCommandArgs>();

		private readonly CompositeControllerFactory CompositeControllerFactory = new CompositeControllerFactory();

		private readonly CommandTable<EditCommandArgs> EditCommandTable = new CommandTable<EditCommandArgs>();
		private readonly CommandTable<GeneralCommandArgs> GeneralCommandTable = new CommandTable<GeneralCommandArgs>();

		private readonly List<HotKey> HotKeys = new List<HotKey>();
		private readonly MessageBoard MessageBoard = new MessageBoard();

		private AllOptions AllOptions;
		private ClipboardManipulator ClipboardManipulator;

		private CompositionContainer CompositionContainer;
		private DirectoryHelper DirectoryHelper;
		private string OptionsFileName;
		private IMessageOutput Output;

		private List<string> Tags;

		public MainForm()
		{
			InitializeComponent();

			SetUp();
		}

		[ImportMany]
		public List<ControllerFactory> ControllerFactories { get; set; }

		private void SetUp()
		{
			Output = new MessageBoardOutput( MessageBoard );

			DirectoryHelper = new DirectoryHelper();
			OptionsFileName = PathHelper.Combine( DirectoryHelper.DataDirectory, @"Options.dat" );

			SetUpCompositionContainer();

			CompositionContainer.ComposeParts( this );

			SetUpCompositeControllerFactory();

			LoadTags();

			LoadOptions();

			ClipboardManipulator = new ClipboardManipulator( Output,
			                                                 CompositionContainer,
			                                                 DirectoryHelper,
			                                                 AllOptions.CommonOptions.ProcessorsPluginsPreference,
			                                                 AllOptions.CommonOptions.FormatSupportersPluginsPreference,
			                                                 AllOptions.CommonOptions.Pasters,
			                                                 ClipboardManipulateCommandTable,
			                                                 EditCommandTable,
			                                                 GeneralCommandTable );

			ShowDummyForm();

			UpdateFromOptions();

			SystemEvents.SessionEnding += SystemEvents_SessionEnding;

			SetUpNotifyIcon1();
		}

		// 最初にFormを表示した時にアクティブにならないバグに対処するおまじない
		private void ShowDummyForm()
		{
			using ( var form = new Form { Size = new Size( 0, 0 ) } )
			{
				form.Show();
				form.Hide();
			}
		}

		private void SetUpNotifyIcon1()
		{
			NotifyIcon1.Icon = (Icon) Icon.Clone();
			NotifyIcon1.Text = "ClipExtension";
			NotifyIcon1.Visible = true;
		}

		private void SetUpCompositeControllerFactory()
		{
			foreach ( ControllerFactory controllerFactory in ControllerFactories )
			{
				CompositeControllerFactory.Add( controllerFactory );
			}

			CompositeControllerFactory.Add( new NoOptionControllerFactory() );
		}

		private void SetUpCompositionContainer()
		{
			var catalog = new AggregateCatalog();

			catalog.Catalogs.Add( new AssemblyCatalog( typeof ( ClipboardManipulator ).Assembly ) );
			catalog.Catalogs.Add( new AssemblyCatalog( typeof ( IPluginMetadata ).Assembly ) );

			catalog.Catalogs.Add( new DirectoryCatalog( DirectoryHelper.PluginsDirectory ) );

			CompositionContainer = new CompositionContainer( catalog );
		}

		private void ShowIncrementalSearchForm()
		{
			ExecuteWithPause( DoShowIncrementalSearchForm );
		}

		private void DoShowIncrementalSearchForm()
		{
			ClipboardManipulator.Repository.Adjust();

			var source = new List<DataUnit>();
			source.AddRange( ClipboardManipulator.Repository.History );
			source.AddRange( ClipboardManipulator.Repository.Favorites );

			using (
				var incrementalSearchForm = new IncrementalSearchForm( Output,
				                                                       ClipboardManipulator,
				                                                       AllOptions.IncrementalSearchFormOptions,
				                                                       source,
				                                                       ClipboardManipulateCommandTable,
				                                                       Tags )
				)
			{
				AllOptions.IncrementalSearchFormOptions.WindowStates.UpdateToForm( incrementalSearchForm );
				incrementalSearchForm.ShowDialog();
				AllOptions.IncrementalSearchFormOptions.WindowStates.UpdateFromForm( incrementalSearchForm );
			}
		}

		private void EnableHotKeys()
		{
			foreach ( HotKey hotKey in HotKeys )
			{
				hotKey.Enabled = true;
			}
		}

		private void DisableHotKeys()
		{
			foreach ( HotKey hotKey in HotKeys )
			{
				hotKey.Enabled = false;
			}
		}

		private void LoadOptions()
		{
			if ( File.Exists( OptionsFileName ) )
			{
				var serializer = new AllOptionsSerializer();

				try
				{
					AllOptions = serializer.DeSerialize( OptionsFileName );
				}
				catch ( Exception e )
				{
					Output.WriteLine( e.Message );
					Output.WriteLine( e.StackTrace );

					AllOptions = new AllOptions();
				}
			}
			else
			{
				AllOptions = new AllOptions();
			}
		}

		private void UpdateFromOptions()
		{
			DisposeHotKeys();
			CreateHotKeys();

			ClipboardManipulator.CheckAndRegisterChainEnabled = AllOptions.CommonOptions.CheckAndRegisterChainEnabled;
			ClipboardManipulator.CheckAndRegisterChainInterval = AllOptions.CommonOptions.CheckAndRegisterChainInterval;
			ClipboardManipulator.CopyReTryDelay = AllOptions.CommonOptions.CopyReTryDelay;
			ClipboardManipulator.CopyReTryTimes = AllOptions.CommonOptions.CopyReTryTimes;
			ClipboardManipulator.BeforePasteDelay = AllOptions.CommonOptions.BeforePasteDelay;
			ClipboardManipulator.Repository.MaxHistoryItemCount = AllOptions.CommonOptions.MaxHistoryItemCount;
		}

		private void VersionToolStripMenuItem_Click( object sender, EventArgs e )
		{
			MessageBox.Show( "1.02" );
		}

		private void ExitToolStripMenuItem_Click( object sender, EventArgs e )
		{
			Exit();
		}

		private void Exit()
		{
			TearDown();
			Application.Exit();
		}

		private void TearDown()
		{
			SystemEvents.SessionEnding -= SystemEvents_SessionEnding;
			DisableHotKeys();

			NotifyIcon1.Visible = false;
			SaveOptions();

			DisposeHotKeys();

			ClipboardManipulator.Dispose();
		}

		private void DisposeHotKeys()
		{
			foreach ( HotKey hotKey in HotKeys )
			{
				hotKey.Dispose();
			}
		}

		private void SaveOptions()
		{
			UpdateToOptions();
			var serializer = new AllOptionsSerializer { ReWriteMode = ReWriteMode.NotEqual, ForceSerialization = true };
			serializer.Serialize( AllOptions, OptionsFileName );
		}

		private void UpdateToOptions()
		{
			AllOptions.CommonOptions.CheckAndRegisterChainEnabled = ClipboardManipulator.CheckAndRegisterChainEnabled;
			AllOptions.CommonOptions.CheckAndRegisterChainInterval = ClipboardManipulator.CheckAndRegisterChainInterval;
			AllOptions.CommonOptions.CopyReTryDelay = ClipboardManipulator.CopyReTryDelay;
			AllOptions.CommonOptions.CopyReTryTimes = ClipboardManipulator.CopyReTryTimes;
			AllOptions.CommonOptions.BeforePasteDelay = ClipboardManipulator.BeforePasteDelay;
			AllOptions.CommonOptions.MaxHistoryItemCount = ClipboardManipulator.Repository.MaxHistoryItemCount;
		}

		private void ShowOptionsForm()
		{
			using ( var optionsForm = new OptionsForm() )
			{
				optionsForm.SetUp( AllOptions, ClipboardManipulator, CompositeControllerFactory, GeneralCommandTable );

				if ( optionsForm.ShowDialog() ==
				     DialogResult.OK )
				{
					UpdateFromOptions();
				}
			}
		}

		private void ShowLogForm()
		{
			using ( var logForm = new LogForm() )
			{
				logForm.SetUp( MessageBoard );

				logForm.ShowDialog();
			}
		}

		private void ShowPluginListForm()
		{
			using ( var form = new PluginListForm() )
			{
				form.SetUp( ClipboardManipulator.Processors, ClipboardManipulator.FormatSupporterUnit.PluginContainer );
				form.ShowDialog();
			}
		}

		private void EditRepository()
		{
			ExecuteWithPause( DoEditRepository );
		}

		private void DoEditRepository()
		{
			ClipboardManipulator.Repository.Adjust();

			using ( var form = new RepositoryEditForm { StartPosition = FormStartPosition.CenterScreen } )
			{
				form.SetUp( ClipboardManipulator.Repository, CompositeControllerFactory, Tags, EditCommandTable, Output );
				form.ShowDialog();
			}
		}

		private void ShowGeneralCommandSearchForm()
		{
			ExecuteWithPause( DoShowGeneralCommandSearchForm );
		}

		private void DoShowGeneralCommandSearchForm()
		{
			ClipboardManipulator.Repository.Adjust();

			using ( var commandSearchForm = new SelectForm { StartPosition = FormStartPosition.CenterScreen, Text = "コマンド" } )
			{
				commandSearchForm.SetSource( GeneralCommandTable.Commands );

				if ( commandSearchForm.ShowDialog() ==
				     DialogResult.OK )
				{
					Thread.Sleep( 50 );

					var selected = (Command<GeneralCommandArgs>) commandSearchForm.SelectedItem;

					ExecuteCommand( selected );
				}
			}
		}

		#region ExecuteCommand

		private Command<GeneralCommandArgs> CurrentCommand;

		private void ExecuteCommandWithPause( Command<GeneralCommandArgs> command )
		{
			CurrentCommand = command;
			ExecuteWithPause( DoExecuteCommand );
		}

		private void ExecuteCommand( Command<GeneralCommandArgs> command )
		{
			CurrentCommand = command;
			DoExecuteCommand();
		}

		private void DoExecuteCommand()
		{
			var args = new GeneralCommandArgs( Output, ClipboardManipulator );
			CurrentCommand.Execute( args );
		}

		#endregion

		#region ExecuteWithPause

		private bool Paused;

		private void ExecuteWithPause( Action action )
		{
			if ( Paused )
			{
				return;
			}

			Paused = true;
			DisableHotKeys();
			ClipboardManipulator.Enabled = false;
			ContextMenuStrip1.Enabled = false;
			try
			{
				action();
			}
			catch ( Exception e )
			{
				MessageBox.Show( e.Message );
				Output.WriteLine( e.Message );
				Output.WriteLine( e.StackTrace );
			}
			finally
			{
				EnableHotKeys();
				ClipboardManipulator.Enabled = true;
				ContextMenuStrip1.Enabled = true;
				Paused = false;
			}
		}

		#endregion

		#region ToolStripMenuItemClick events

		private void GeneralCommandsToolStripMenuItem_Click( object sender, EventArgs e )
		{
			ShowGeneralCommandSearchForm();
		}

		private void SystemEvents_SessionEnding( object sender, SessionEndingEventArgs e )
		{
			Exit();
		}

		private void IncrementalSearchToolStripMenuItem_Click( object sender, EventArgs e )
		{
			ShowIncrementalSearchForm();
		}

		private void CheckAndRegisterChainToolStripMenuItem_Click( object sender, EventArgs e )
		{
			ClipboardManipulator.CheckAndRegisterChain();
		}

		private void OptionsToolStripMenuItem_Click( object sender, EventArgs e )
		{
			ExecuteWithPause( ShowOptionsForm );
		}

		private void LogToolStripMenuItem_Click( object sender, EventArgs e )
		{
			ExecuteWithPause( ShowLogForm );
		}

		private void PluginsToolStripMenuItem_Click( object sender, EventArgs e )
		{
			ExecuteWithPause( ShowPluginListForm );
		}

		private void EditToolStripMenuItem_Click( object sender, EventArgs e )
		{
			EditRepository();
		}

		private void SaveRepositoryToolStripMenuItem_Click( object sender, EventArgs e )
		{
			ExecuteWithPause( ClipboardManipulator.SaveRepository );
		}

		#endregion

		#region HotKey Pressed events

		private void ShowRepositoryEditFormKey_Pressed( object sender, EventArgs e )
		{
			EditRepository();
		}

		private void ShowIncrementalSearchFormKey_Pressed( object sender, EventArgs e )
		{
			ShowIncrementalSearchForm();
		}

		private void ShowGeneralCommandSearchFormKey_Pressed( object sender, EventArgs e )
		{
			ShowGeneralCommandSearchForm();
		}

		private void GeneralCommandHotKeyPressed( object sender, EventArgs e )
		{
			object tag = ( (HotKey) sender ).Tag;
			ExecuteCommandWithPause( (Command<GeneralCommandArgs>) tag );
		}

		#endregion

		#region CreateHotKeys

		private void CreateHotKeys()
		{
			HotKey showIncrementalSearchFormKey = CreateHotKey();
			showIncrementalSearchFormKey.CombinationKeys = AllOptions.CommonOptions.ShowIncrementalSearchFormKey;
			showIncrementalSearchFormKey.Pressed += ShowIncrementalSearchFormKey_Pressed;

			HotKey showRepositoryEditFormKey = CreateHotKey();
			showRepositoryEditFormKey.CombinationKeys = AllOptions.CommonOptions.ShowRepositoryEditFormKey;
			showRepositoryEditFormKey.Pressed += ShowRepositoryEditFormKey_Pressed;

			HotKey showGeneralCommandSearchFormKey = CreateHotKey();
			showGeneralCommandSearchFormKey.CombinationKeys = AllOptions.CommonOptions.ShowGeneralCommandSearchFormKey;
			showGeneralCommandSearchFormKey.Pressed += ShowGeneralCommandSearchFormKey_Pressed;

			foreach ( var pair in AllOptions.CommonOptions.GeneralCommandKeys )
			{
				foreach ( var command in GeneralCommandTable.Commands )
				{
					if ( pair.Key ==
					     command.Name )
					{
						HotKey hotKey = CreateHotKey();
						hotKey.CombinationKeys = pair.Value;
						hotKey.Tag = command;
						hotKey.Pressed += GeneralCommandHotKeyPressed;

						break;
					}
				}
			}
		}

		private HotKey CreateHotKey()
		{
			var result = new HotKey();
			HotKeys.Add( result );
			return result;
		}

		#endregion

		#region LoadTags

		private void LoadTags()
		{
			string fileName = PathHelper.Combine( DirectoryHelper.DataDirectory, @"Tags.txt" );

			if ( !File.Exists( fileName ) )
			{
				var defaultTags = new StringList();
				defaultTags.AddRange( new[] { "AA", "Mail", "Path", "URL", "Net", "Business" } );
				defaultTags.Sort();
				defaultTags.SaveToFile( fileName );
			}

			var list = new StringList( fileName );
			Tags = new List<string>();

			foreach ( string item in list )
			{
				string trimmed = item.Trim();

				if ( trimmed != string.Empty &&
				     !Tags.Contains( trimmed ) )
				{
					Tags.Add( trimmed );
				}
			}
		}

		#endregion
	}
}