﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using ClipExtension.Common;
using ClipExtension.Controls;
using ClipExtension.Main;
using ClipExtension.Plugins;
using Tomham;
using Tomham.Collections.Generic;
using Tomham.Windows.Forms;
using Tomham.Windows.Forms.AutoGeneration;

namespace ClipExtension
{
	public partial class RepositoryEditForm : Form
	{
		private ControllerFactory ControllerFactory;
		private CommandTable<EditCommandArgs> EditCommandTable;
		private IMessageOutput Output;
		private List<string> Tags;

		public RepositoryEditForm()
		{
			InitializeComponent();
		}

		public Repository Target { get; private set; }

		public void SetUp( Repository target,
		                   ControllerFactory factory,
		                   List<string> tags,
		                   CommandTable<EditCommandArgs> editCommandTable,
		                   IMessageOutput output )
		{
			Target = target;
			ControllerFactory = factory;
			Tags = tags;
			EditCommandTable = editCommandTable;
			Output = output;

			HistoryDataUnitsViewer.ContentsFilterListBox.SetSource( Target.History );
			HistoryDataUnitsViewer.ContentsFilterListBox.ListBoxKeyDown += HistoryContentsFilterListBox_ListBoxKeyDown;
			HistoryDataUnitsViewer.ContentsFilterListBox.ListBoxDoubleClick += HistoryContentsFilterListBox_ListBoxDoubleClick;
			HistoryDataUnitsViewer.ContentsFilterListBox.TextBoxKeyDown += HistoryContentsOrTagsTextBoxKeyDown;
			HistoryDataUnitsViewer.TagsLabeledComboBox.ComboBox.KeyDown += HistoryContentsOrTagsTextBoxKeyDown;
			HistoryDataUnitsViewer.TagsLabeledComboBox.ComboBox.Items.AddRange( tags.ToArray() );

			FavoritesDataUnitsViewer.ContentsFilterListBox.SetSource( Target.Favorites );
			FavoritesDataUnitsViewer.ContentsFilterListBox.ListBoxKeyDown += FavoritesContentsFilterListBox_ListBoxKeyDown;
			FavoritesDataUnitsViewer.ContentsFilterListBox.ListBoxDoubleClick +=
				FavoritesContentsFilterListBox_ListBoxDoubleClick;
			FavoritesDataUnitsViewer.ContentsFilterListBox.TextBoxKeyDown += FavoritesContentsOrTagsTextBoxKeyDown;
			FavoritesDataUnitsViewer.TagsLabeledComboBox.ComboBox.KeyDown += FavoritesContentsOrTagsTextBoxKeyDown;
			FavoritesDataUnitsViewer.TagsLabeledComboBox.ComboBox.Items.AddRange( tags.ToArray() );
		}

		private void FavoritesContentsOrTagsTextBoxKeyDown( object sender, KeyEventArgs e )
		{
			if ( e.KeyData ==
			     Keys.Enter )
			{
				e.SuppressKeyPress = true;
				EditFavoritesSelectedItem();
			}
		}

		private void HistoryContentsOrTagsTextBoxKeyDown( object sender, KeyEventArgs e )
		{
			if ( e.KeyData ==
			     Keys.Enter )
			{
				e.SuppressKeyPress = true;
				EditHistorySelectedItem();
			}
		}

		private void FavoritesContentsFilterListBox_ListBoxKeyDown( object sender, KeyEventArgs e )
		{
			switch ( e.KeyData )
			{
				case Keys.Delete:
					e.SuppressKeyPress = true;
					RemoveFavoritesSelectedItem();
					break;

				case Keys.Enter:
					e.SuppressKeyPress = true;
					EditFavoritesSelectedItem();
					break;
			}
		}

		private void HistoryContentsFilterListBox_ListBoxKeyDown( object sender, KeyEventArgs e )
		{
			switch ( e.KeyData )
			{
				case Keys.Delete:
					e.SuppressKeyPress = true;
					RemoveHistorySelectedItems();
					break;

				case Keys.Enter:
					e.SuppressKeyPress = true;
					EditHistorySelectedItem();
					break;
			}
		}

		private void FavoritesContentsFilterListBox_ListBoxDoubleClick( object sender, EventArgs e )
		{
			EditFavoritesSelectedItem();
		}

		private void HistoryContentsFilterListBox_ListBoxDoubleClick( object sender, EventArgs e )
		{
			EditHistorySelectedItem();
		}

		private void UpdateToControls()
		{
			HistoryDataUnitsViewer.ContentsFilterListBox.UpdateMatchedItems();
			FavoritesDataUnitsViewer.ContentsFilterListBox.UpdateMatchedItems();
		}

		private void MoveSelectedItems( List<DataUnit> selectedItems, List<DataUnit> from, List<DataUnit> to )
		{
			to.AddRange( selectedItems );

			RemoveSelectedItems( selectedItems, from );
		}

		private void EditSelectedItem( DataUnitsViewer dataUnitsViewer )
		{
			var selectedItem = (DataUnit) dataUnitsViewer.ContentsFilterListBox.GetOnlyOneSelectedItem();

			if ( selectedItem == null )
			{
				return;
			}

			using ( var form = new DataUnitEditForm() )
			{
				form.SetUp( selectedItem, ControllerFactory, Tags );

				if ( form.ShowDialog() ==
				     DialogResult.OK )
				{
					UpdateToControls();
				}
			}
		}

		private void RemoveSelectedItems( IEnumerable<DataUnit> selectedItems, List<DataUnit> list )
		{
			foreach ( DataUnit selectedItem in selectedItems )
			{
				list.RemoveReference( selectedItem );
			}
		}

		private void EditHistorySelectedItem()
		{
			EditSelectedItem( HistoryDataUnitsViewer );
		}

		private void RemoveHistorySelectedItems()
		{
			if ( !HistoryDataUnitsViewer.ContentsFilterListBox.HasSelectedItems() )
			{
				return;
			}

			RemoveSelectedItems( HistoryDataUnitsViewer.ContentsFilterListBox.GetSelectedItems<DataUnit>(), Target.History );
			UpdateToControls();
		}

		private void MoveToFavorites()
		{
			if ( !HistoryDataUnitsViewer.ContentsFilterListBox.HasSelectedItems() )
			{
				return;
			}

			MoveSelectedItems( HistoryDataUnitsViewer.ContentsFilterListBox.GetSelectedItems<DataUnit>(),
			                   Target.History,
			                   Target.Favorites );
			UpdateToControls();
		}

		private void MoveToHistory()
		{
			if ( !FavoritesDataUnitsViewer.ContentsFilterListBox.HasSelectedItems() )
			{
				return;
			}

			MoveSelectedItems( FavoritesDataUnitsViewer.ContentsFilterListBox.GetSelectedItems<DataUnit>(),
			                   Target.Favorites,
			                   Target.History );
			UpdateToControls();
		}

		private void EditFavoritesSelectedItem()
		{
			EditSelectedItem( FavoritesDataUnitsViewer );
		}

		private void RemoveFavoritesSelectedItem()
		{
			if ( !FavoritesDataUnitsViewer.ContentsFilterListBox.HasSelectedItems() )
			{
				return;
			}

			RemoveSelectedItems( FavoritesDataUnitsViewer.ContentsFilterListBox.GetSelectedItems<DataUnit>(), Target.Favorites );
			UpdateToControls();
		}

		private void ShowCommandSearchForm( DataUnitsViewer dataUnitsViewer, bool targetsFromHistory )
		{
			List<DataUnit> dataUnits = dataUnitsViewer.ContentsFilterListBox.GetSelectedItems<DataUnit>();

			using ( var commandSearchForm = new SelectForm { Text = "コマンド" } )
			{
				commandSearchForm.SetSource( EditCommandTable.Commands );

				if ( commandSearchForm.ShowDialog() ==
				     DialogResult.OK )
				{
					var args = new EditCommandArgs( Output,
					                                Target,
					                                dataUnits,
					                                targetsFromHistory );

					var selected = (Command<EditCommandArgs>) commandSearchForm.SelectedItem;

					try
					{
						selected.Execute( args );
					}
					catch ( Exception e )
					{
						MessageBox.Show( e.Message );
						Output.WriteLine( e.Message );
						Output.WriteLine( e.StackTrace );
					}

					UpdateToControls();
				}
			}
		}

		private void RepositoryEditForm_Shown( object sender, EventArgs e )
		{
			HistoryDataUnitsViewer.ContentsFilterListBox.FocusToTextBox();
		}

		private void AddTag( DataUnit dataUnit, string tag )
		{
			if ( !dataUnit.Tags.Contains( tag ) )
			{
				dataUnit.Tags.Add( tag );
			}
		}

		private void RemoveTag( DataUnit dataUnit, string tag )
		{
			if ( dataUnit.Tags.Contains( tag ) )
			{
				dataUnit.Tags.Remove( tag );
			}
		}

		#region SelectTagsAction

		private void SelectTagsAction( DataUnitsViewer dataUnitsViewer, Action<DataUnit, string> action )
		{
			List<DataUnit> dataUnits = dataUnitsViewer.ContentsFilterListBox.GetSelectedItems<DataUnit>();
			if ( dataUnits.Count == 0 )
			{
				return;
			}

			IEnumerable<string> selectedTags = SelectTags();

			if ( selectedTags == null )
			{
				return;
			}

			foreach ( DataUnit dataUnit in dataUnits )
			{
				foreach ( string tag in selectedTags )
				{
					action( dataUnit, tag );
				}
			}

			UpdateToControls();
		}

		private IEnumerable<string> SelectTags()
		{
			using ( var form = new SelectForm { Text = "タグ選択", SelectionMode = SelectionMode.MultiExtended } )
			{
				form.SetSource( Tags );

				if ( form.ShowDialog() ==
				     DialogResult.OK )
				{
					return form.SelectedItems.Cast<string>();
				}
			}

			return null;
		}

		#endregion

		#region ToolStripButton Click events

		private void toolStripButtonMoveToFavorites_Click( object sender, EventArgs e )
		{
			MoveToFavorites();
		}

		private void toolStripButtonHistoryEditSelectedItem_Click( object sender, EventArgs e )
		{
			EditHistorySelectedItem();
		}

		private void toolStripButtonHistoryRemoveSelectedItems_Click( object sender, EventArgs e )
		{
			RemoveHistorySelectedItems();
		}

		private void toolStripButtonFavoritesEditSelectedItem_Click( object sender, EventArgs e )
		{
			EditFavoritesSelectedItem();
		}

		private void toolStripButtonFavoritesRemoveSelectedItems_Click( object sender, EventArgs e )
		{
			RemoveFavoritesSelectedItem();
		}

		private void toolStripButtonHistoryShowCommandSearchForm_Click( object sender, EventArgs e )
		{
			ShowCommandSearchForm( HistoryDataUnitsViewer, true );
		}

		private void toolStripButtonFavoriteShowCommandSearchForm_Click( object sender, EventArgs e )
		{
			ShowCommandSearchForm( FavoritesDataUnitsViewer, false );
		}

		private void toolStripButtonMoveToHistory_Click( object sender, EventArgs e )
		{
			MoveToHistory();
		}

		private void toolStripButtonHistoryAddTags_Click( object sender, EventArgs e )
		{
			SelectTagsAction( HistoryDataUnitsViewer, AddTag );
		}

		private void toolStripButtonHistoryRemoveTags_Click( object sender, EventArgs e )
		{
			SelectTagsAction( HistoryDataUnitsViewer, RemoveTag );
		}

		private void toolStripButtonFavoritesAddTags_Click( object sender, EventArgs e )
		{
			SelectTagsAction( FavoritesDataUnitsViewer, AddTag );
		}

		private void toolStripButtonFavoritesRemoveTags_Click( object sender, EventArgs e )
		{
			SelectTagsAction( FavoritesDataUnitsViewer, RemoveTag );
		}

		#endregion
	}
}