﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Tomham.Collections.Generic
{
	public static class CollectionHelper
	{
		public static void MoveUpItems<T>( this IList<T> list, params int[] targetIndices )
		{
			int excludeToIndex = -1;

			foreach ( int fromIndex in targetIndices )
			{
				Move( fromIndex, fromIndex - 1, list, ref excludeToIndex );
			}
		}

		private static void Move<T>( int fromIndex, int toIndex, IList<T> list, ref int excludeToIndex )
		{
			if ( HasToNeedToMove( toIndex, list, excludeToIndex ) )
			{
				SwapItems( list, fromIndex, toIndex );
			}
			else
			{
				excludeToIndex = fromIndex;
			}
		}

		public static void SwapItems<T>( this IList<T> list, int index1, int index2 )
		{
			T object1 = list[index1];
			T object2 = list[index2];

			list[index2] = object1;
			list[index1] = object2;
		}

		private static bool HasToNeedToMove<T>( int toIndex, IList<T> list, int excludeToIndex )
		{
			if ( !ExistsIndex( list, toIndex ) )
			{
				return false;
			}

			if ( toIndex == excludeToIndex )
			{
				return false;
			}

			return true;
		}

		public static bool ExistsIndex<T>( this IList<T> list, int index )
		{
			if ( index < 0 )
			{
				return false;
			}

			if ( list.Count <= index )
			{
				return false;
			}

			return true;
		}

		public static bool IsLastIndex<T>( this List<T> list, int index )
		{
			if ( ExistsIndex( list, index ) )
			{
				return list.Count - 1 == index;
			}

			return false;
		}

		public static void MoveDownItems<T>( this IList<T> list, params int[] targetIndices )
		{
			int excludeToIndex = -1;

			for ( int i = targetIndices.Length - 1; 0 <= i; i-- )
			{
				int fromIndex = targetIndices[i];

				Move( fromIndex, fromIndex + 1, list, ref excludeToIndex );
			}
		}

		public static void MoveItemToFirst<T>( this IList<T> list, T item )
		{
			if ( !list.Contains( item ) )
			{
				throw new ArgumentException( "Not contains item." );
			}

			list.Remove( item );
			list.Insert( 0, item );
		}

		public static void MoveItemToFirst<T>( this IList<T> list, int index )
		{
			T item = list[index];

			list.RemoveAt( index );
			list.Insert( 0, item );
		}

		public static void MoveItemsToFirst<T>( this IList<T> list, IList<T> targets )
		{
			IEnumerable<T> intersect = list.Intersect( targets );
			var comparer = new EnumerableEqualityComparer<T>();
			if ( !comparer.Equals( targets, intersect ) )
			{
				throw new ArgumentException();
			}

			for ( int i = targets.Count - 1; i >= 0; i-- )
			{
				list.MoveItemToFirst( targets[i] );
			}
		}

		public static void RemoveExcessItems<T>( this IList<T> list, int maxCount )
		{
			if ( maxCount < 0 )
			{
				throw new ArgumentOutOfRangeException( "maxCount" );
			}

			if ( maxCount < list.Count )
			{
				RemoveRange( list, maxCount, list.Count - maxCount );
			}
		}

		public static void AddRange<T>( ICollection<T> to, IEnumerable<T> addition )
		{
			if ( to == null )
			{
				throw new ArgumentNullException( "to" );
			}
			if ( addition == null )
			{
				throw new ArgumentNullException( "addition" );
			}

			foreach ( T item in addition )
			{
				to.Add( item );
			}
		}

		public static void RemoveRange<T>( IList<T> list, int index, int count )
		{
			if ( index < 0 )
			{
				throw new ArgumentOutOfRangeException( "index" );
			}

			if ( count < 0 )
			{
				throw new ArgumentOutOfRangeException( "count" );
			}

			if ( count == 0 )
			{
				return;
			}

			if ( !list.ExistsIndex( index + count - 1 ) )
			{
				throw new ArgumentException();
			}

			for ( int i = 0; i < count; i++ )
			{
				list.RemoveAt( index );
			}
		}

		public static void ClearAndAddRange<T>( ICollection<T> to, IEnumerable<T> addition )
		{
			to.Clear();
			AddRange( to, addition );
		}

		public static void RemoveDuplicateItems<T>( this IList<T> list )
		{
			list.RemoveDuplicateItems( EqualityComparer<T>.Default );
		}

		public static void RemoveDuplicateItems<T>( this IList<T> list, EqualityComparer<T> comparer )
		{
			for ( int currentIndex = 0; currentIndex < list.Count; currentIndex++ )
			{
				var removeIndexes = new List<int>();

				for ( int otherIndex = currentIndex + 1; otherIndex < list.Count; otherIndex++ )
				{
					if ( comparer.Equals( list[currentIndex], list[otherIndex] ) )
					{
						removeIndexes.Add( otherIndex );
					}
				}

				for ( int i = removeIndexes.Count - 1; i >= 0; i-- )
				{
					list.RemoveAt( removeIndexes[i] );
				}
			}
		}

		public static bool RemoveReference<T>( this IList<T> list, object item )
		{
			if ( !list.ContainsReference( item ) )
			{
				return false;
			}

			int index = list.ReferenceIndexOf( item );

			list.RemoveAt( index );

			return true;
		}

		public static int ReferenceIndexOf<T>( this IList<T> list, object item )
		{
			for ( int index = 0; index < list.Count; index++ )
			{
				T current = list[index];

				if ( ReferenceEquals( current, item ) )
				{
					return index;
				}
			}

			return -1;
		}
	}
}