using System;
using System.Collections.Generic;
using System.IO;
using NUnit.Extensions.Converter;
using NUnit.Framework;

namespace Tomham.IO.Tests
{
	[TestFixture]
	public class PathHelperTest : AssertionHelper
	{
		[Test]
		public void Combine()
		{
			string result = PathHelper.Combine( @"D:", @"Sample.txt" );

			Assert.AreEqual( @"D:\Sample.txt", result );

			result = PathHelper.Combine( @"D:\", @"Sample.txt" );

			Assert.AreEqual( @"D:\Sample.txt", result );

			result = PathHelper.Combine( @"D:\Work", @"Sample.txt" );

			Assert.AreEqual( @"D:\Work\Sample.txt", result );
		}

		[Test]
		public void ExcludeTrailingPathDelimiter1()
		{
			string path = @"C:\Windows";
			string rightPath = @"C:\Windows";

			string result = PathHelper.ExcludeTrailingPathDelimiter( path );

			Console.WriteLine( result );
			Assert.IsTrue( result == rightPath );
		}

		[Test]
		public void ExcludeTrailingPathDelimiter2()
		{
			string path = @"C:\Windows\";
			string rightPath = @"C:\Windows";

			string result = PathHelper.ExcludeTrailingPathDelimiter( path );

			Console.WriteLine( result );
			Assert.IsTrue( result == rightPath );
		}

		[Test]
		public void GetDrive()
		{
			string actual = PathHelper.GetDrive( @"V:\Work" );

			Expect( actual, Is.EqualTo( @"V:" ) );

			actual = PathHelper.GetDrive( @"V:\" );

			Expect( actual, Is.EqualTo( @"V:" ) );

			actual = PathHelper.GetDrive( @"V:" );

			Expect( actual, Is.EqualTo( @"V:" ) );
		}

		[Test]
		[ExpectedException( typeof ( ArgumentException ) )]
		public void GetDriveWithEmptyPath()
		{
			PathHelper.GetDrive( string.Empty );
		}

		[Test]
		[ExpectedException( typeof ( ArgumentException ) )]
		public void GetDriveWithoutDrivePattern()
		{
			PathHelper.GetDrive( "Sample" );
		}

		[Test]
		[ExpectedException( typeof ( ArgumentException ) )]
		public void GetDriveWithTooShortPath1()
		{
			PathHelper.GetDrive( "V" );
		}

		[Test]
		public void GetName()
		{
			Assert.AreEqual( @"C:", PathHelper.GetName( @"C:" ) );
			Assert.AreEqual( @"C:", PathHelper.GetName( @"C:\" ) );
			Assert.AreEqual( @"Dir", PathHelper.GetName( @"C:\Dir" ) );
			Assert.AreEqual( @"File.txt", PathHelper.GetName( @"C:\Dir\File.txt" ) );
		}

		[Test]
		public void GetParentWithUnExistsDirectory()
		{
			string result = PathHelper.GetParent( @"V:\Test" );

			Expect( result == @"V:" );

			result = PathHelper.GetParent( @"V:\Test\Child" );

			Expect( result == @"V:\Test" );
		}

		[Test]
		public void GetParentWithUnExistsDrive()
		{
			string result = PathHelper.GetParent( @"V:" );

			Expect( result == string.Empty );

			result = PathHelper.GetParent( @"V:\" );

			Expect( result == string.Empty );
		}

		[Test]
		public void GetParentWithUnExistsFile()
		{
			string result = PathHelper.GetParent( @"V:\Test\Sample.txt" );

			Expect( result == @"V:\Test" );

			result = PathHelper.GetParent( @"V:\Sample.txt" );

			Expect( result == @"V:" );
		}

		[Test]
		public void GetUnExistsDirectoriesWithExistsDirectories()
		{
			Converter converter = new DefaultConverter();

			object[] expected = new object[] { };

			IList<string> result = PathHelper.GetUnExistsDirectories( @"J:\Soft" );

			converter.AssertEquals( expected, result );
		}

		[Test]
		public void GetUnExistsDirectoriesWithExistsDrive()
		{
			Converter converter = new DefaultConverter();

			object[] expected = new object[] { };

			IList<string> result = PathHelper.GetUnExistsDirectories( @"J:\" );

			converter.AssertEquals( expected, result );
		}

		[Test]
		public void GetUnExistsDirectoriesWithUnExistsDirectories()
		{
			Converter converter = new DefaultConverter();

			object[] expected = new object[] { @"J:\UnExists", @"J:\UnExists\A", @"J:\UnExists\A\B" };

			IList<string> result = PathHelper.GetUnExistsDirectories( @"J:\UnExists\A\B" );

			converter.AssertEquals( expected, result );

			result = PathHelper.GetUnExistsDirectories( @"J:\UnExists\A\B\" );

			converter.AssertEquals( expected, result );
		}

		[Test]
		[ExpectedException( typeof ( DriveNotFoundException ) )]
		public void GetUnExistsDirectoriesWithUnExistsDrivePath()
		{
			PathHelper.GetUnExistsDirectories( @"X:\UnExists\A\B" );
		}

		[Test]
		public void IncludeTrailingPathDelimiter1()
		{
			string path = @"C:\Windows";
			string rightPath = @"C:\Windows\";

			string result = PathHelper.IncludeTrailingPathDelimiter( path );

			Console.WriteLine( result );
			Assert.IsTrue( result == rightPath );
		}

		[Test]
		public void IncludeTrailingPathDelimiter2()
		{
			string path = @"C:\Windows\";
			string rightPath = @"C:\Windows\";

			string result = PathHelper.IncludeTrailingPathDelimiter( path );

			Console.WriteLine( result );
			Assert.IsTrue( result == rightPath );
		}

		[Test]
		public void IsDriveCaseOfDrive1()
		{
			Assert.IsTrue( PathHelper.IsDrive( @"C:" ) );
		}

		[Test]
		public void IsDriveCaseOfDrive2()
		{
			Assert.IsTrue( PathHelper.IsDrive( @"C:\" ) );
		}

		[Test]
		public void IsDriveCaseOfEmpty()
		{
			Assert.IsFalse( PathHelper.IsDrive( string.Empty ) );
		}

		[Test]
		public void IsDriveCaseOfFileName()
		{
			Assert.IsFalse( PathHelper.IsDrive( @"J:\soft\irvine\readme.txt" ) );
		}
	}
}