﻿using System.Collections.Generic;
using System.Windows.Forms;
using NUnit.Framework;

namespace Tomham.Windows.Forms.Tests
{
	[TestFixture]
	public class ListBoxHelperTest
	{
		private ListBox CreateListBox()
		{
			var listBox = new ListBox();

			listBox.BeginUpdate();

			listBox.Items.AddRange(
				new object[] { "ant", "*bird", "cat", "*dog", "*eagle", "fly" } );

			listBox.SetSelected( 1, true );
			listBox.SetSelected( 3, true );
			listBox.SetSelected( 4, true );

			listBox.EndUpdate();

			return listBox;
		}

		[Test]
		public void MoveDownSelectedItems()
		{
			ListBox actual = CreateListBox();

			actual.MoveDownSelectedItems();
			object expected = new List<string> { "ant", "cat", "*bird", "fly", "*dog", "*eagle" };
			Assert.That( actual.Items, Is.EqualTo( expected ) );

			actual.MoveDownSelectedItems();
			expected = new List<string> { "ant", "cat", "fly", "*bird", "*dog", "*eagle" };
			Assert.That( actual.Items, Is.EqualTo( expected ) );

			actual.MoveDownSelectedItems();

			Assert.That( actual.Items, Is.EqualTo( expected ) );
		}

		[Test]
		public void MoveSelectedItemsToOtherWithSelectedItems()
		{
			ListBox one = CreateListBox();
			var other = new ListBox();
			other.Items.Add( "giraffe" );
			other.Items.Add( "high" );
			other.SetSelected( 1, true );

			one.MoveSelectedItemsToOther( other );

			object expectedOne = new List<string> { "ant", "cat", "fly" };

			object expectedOther = new List<string> { "giraffe", "*bird", "*dog", "*eagle", "high" };

			Assert.That( one.Items, Is.EqualTo( expectedOne ) );
			Assert.That( other.Items, Is.EqualTo( expectedOther ) );

			object expectedSelectedIndices = new List<int> { 1, 2, 3 };
			Assert.That( other.SelectedIndices, Is.EqualTo( expectedSelectedIndices ) );

			one.MoveSelectedItemsToOther( other );
			Assert.That( one.Items, Is.EqualTo( expectedOne ) );
			Assert.That( other.Items, Is.EqualTo( expectedOther ) );
			Assert.That( other.SelectedIndices, Is.EqualTo( expectedSelectedIndices ) );
		}

		[Test]
		public void MoveSelectedItemsToOtherWithoutSelectedItems()
		{
			ListBox one = CreateListBox();
			var other = new ListBox();
			other.Items.Add( "giraffe" );
			other.Items.Add( "high" );

			one.MoveSelectedItemsToOther( other );

			object expectedOne = new List<string> { "ant", "cat", "fly" };

			object expectedOther = new List<string> { "giraffe", "high", "*bird", "*dog", "*eagle" };

			Assert.That( one.Items, Is.EqualTo( expectedOne ) );
			Assert.That( other.Items, Is.EqualTo( expectedOther ) );

			object expectedSelectedIndices = new List<int> { 2, 3, 4 };
			Assert.That( other.SelectedIndices, Is.EqualTo( expectedSelectedIndices ) );

			one.MoveSelectedItemsToOther( other );
			Assert.That( one.Items, Is.EqualTo( expectedOne ) );
			Assert.That( other.Items, Is.EqualTo( expectedOther ) );
			Assert.That( other.SelectedIndices, Is.EqualTo( expectedSelectedIndices ) );
		}

		[Test]
		public void MoveUpSelectedItems()
		{
			ListBox actual = CreateListBox();

			actual.MoveUpSelectedItems();
			object expected = new List<string> { "*bird", "ant", "*dog", "*eagle", "cat", "fly" };
			Assert.That( actual.Items, Is.EqualTo( expected ) );

			actual.MoveUpSelectedItems();
			expected = new List<string> { "*bird", "*dog", "*eagle", "ant", "cat", "fly" };
			Assert.That( actual.Items, Is.EqualTo( expected ) );

			actual.MoveUpSelectedItems();

			Assert.That( actual.Items, Is.EqualTo( expected ) );
		}

		[Test]
		public void MoveUpSelectedItemsWithNoItem()
		{
			var noItemListBox = new ListBox();

			noItemListBox.MoveUpSelectedItems();
			Assert.That( noItemListBox.Items.Count, Is.EqualTo( 0 ) );
		}

		[Test]
		public void ReplaceItem()
		{
			var actual = new ListBox();
			actual.Items.AddRange( new object[] { "1", "2", "3" } );
			actual.SetSelected( 1, true );

			actual.ReplaceItem( 0, "Replaced1" );

			object expected = new List<string> { "Replaced1", "2", "3" };

			Assert.That( actual.Items, Is.EqualTo( expected ) );
			Assert.That( actual.SelectedIndex, Is.EqualTo( 1 ) );

			actual.ReplaceItem( 1, "Replaced2" );

			expected = new List<string> { "Replaced1", "Replaced2", "3" };

			Assert.That( actual.Items, Is.EqualTo( expected ) );
			Assert.That( actual.SelectedIndex, Is.EqualTo( 1 ) );
		}

		[Test]
		public void Reverse()
		{
			var actual = new ListBox();
			actual.Items.AddRange( new object[] { "*ant", "*bird", "cat", "dog" } );
			actual.SetSelected( 0, true );
			actual.SetSelected( 1, true );

			actual.Reverse();
			object expected = new List<string> { "dog", "cat", "*bird", "*ant" };
			Assert.That( actual.Items, Is.EqualTo( expected ) );

			expected = new List<string> { "*bird", "*ant" };
			Assert.That( actual.SelectedItems, Is.EqualTo( expected ) );
		}

		[Test]
		public void SwapItems()
		{
			ListBox actual = CreateListBox();

			actual.SwapItems( 0, 4 );

			object expected = new List<string> { "*eagle", "*bird", "cat", "*dog", "ant", "fly" };

			Assert.That( actual.Items, Is.EqualTo( expected ) );

			actual.SwapItems( 0, 5 );

			expected = new List<string> { "fly", "*bird", "cat", "*dog", "ant", "*eagle" };

			Assert.That( actual.Items, Is.EqualTo( expected ) );
		}
	}
}