﻿using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Tomham.Windows.Forms
{
	/// <summary>
	/// OSのクリップボードチェーンに参加して監視します
	/// </summary>
	/// <remarks>必ずDisposeを呼んで下さい。デストラクタではエラーが出るので付いてません</remarks>
	public class ClipboardWatcher : IDisposable
	{
		private readonly Timer CheckAndRegisterChainTimer;
		private readonly ClipboardWatchForm WatchForm;

		public ClipboardWatcher()
		{
			WatchForm = new ClipboardWatchForm( this );
			CheckAndRegisterChainTimer = new Timer();
			CheckAndRegisterChainTimer.Tick += CheckAndRegisterChainTimerTick;

			Enabled = true;
			CheckAndRegisterChainInterval = 1000 * 10;
		}

		public bool Enabled { get; set; }

		public bool CheckAndRegisterChainEnabled
		{
			get { return CheckAndRegisterChainTimer.Enabled; }
			set { CheckAndRegisterChainTimer.Enabled = value; }
		}

		public int CheckAndRegisterChainInterval
		{
			get { return CheckAndRegisterChainTimer.Interval; }
			set { CheckAndRegisterChainTimer.Interval = value; }
		}

		#region IDisposable Members

		public void Dispose()
		{
			CheckAndRegisterChainTimer.Dispose();
			WatchForm.Dispose();
		}

		#endregion

		public event EventHandler<DrawClipboardEventArgs> DrawClipboard;

		protected virtual void OnDrawClipboard( DrawClipboardEventArgs e )
		{
			if ( DrawClipboard != null )
			{
				DrawClipboard( this, e );
			}
		}

		public void CheckAndRegisterChain()
		{
			WatchForm.CheckAndRegisterChain();
		}

		private void CheckAndRegisterChainTimerTick( object sender, EventArgs e )
		{
			CheckAndRegisterChain();
		}

		#region Nested type: ClipboardWatchForm

		private class ClipboardWatchForm : Form
		{
			private readonly ClipboardWatcher ClipboardWatcher;

			private bool AlreadyDisposed;
			private IntPtr NextHandleStore = IntPtr.Zero;

			public ClipboardWatchForm( ClipboardWatcher clipboardWatcher )
			{
				ClipboardWatcher = clipboardWatcher;
				CheckAndRegisterChain();
			}

			private IntPtr NextHandle
			{
				get { return NextHandleStore; }
				set
				{
					if ( IsMyHandle( value ) )
					{
						return;
					}

					NextHandleStore = value;
				}
			}

			protected override void Dispose( bool disposing )
			{
				if ( !AlreadyDisposed )
				{
					UnRegisterChain();
					AlreadyDisposed = true;
				}
				base.Dispose( disposing );
			}

			[DllImport( "user32.dll" )]
			private static extern IntPtr GetClipboardViewer();

			[DllImport( "user32.dll" )]
			private static extern IntPtr SetClipboardViewer( IntPtr hWndNewViewer );

			[DllImport( "user32.dll" )]
			private static extern bool ChangeClipboardChain( IntPtr hWndRemove, IntPtr hWndNewNext );

			[DllImport( "user32.dll", CharSet = CharSet.Auto )]
			private static extern int SendMessage( IntPtr hWnd, int msg, IntPtr wParam, IntPtr lParam );

			private bool IsMyHandle( IntPtr handle )
			{
				return handle == Handle;
			}

			private void ClearNextHandle()
			{
				NextHandle = IntPtr.Zero;
			}

			public void CheckAndRegisterChain()
			{
				if ( AlreadyDisposed )
				{
					throw new ObjectDisposedException( "ClipboardWatchForm" );
				}

				if ( IsMyHandle( GetClipboardViewer() ) )
				{
					return;
				}

				UnRegisterChain();
				NextHandle = SetClipboardViewer( Handle );
			}

			private void UnRegisterChain()
			{
				ChangeClipboardChain( Handle, NextHandle );
			}

			protected override void WndProc( ref Message m )
			{
				switch ( m.Msg )
				{
					case WindowMessage.DRAWCLIPBOARD:

						try
						{
							if ( ClipboardWatcher.Enabled )
							{
								ClipboardWatcher.OnDrawClipboard( new DrawClipboardEventArgs() );
							}
						}
						finally
						{
							if ( NextHandle != IntPtr.Zero )
							{
								SendMessage( NextHandle, m.Msg, m.WParam, m.LParam );
							}
						}

						break;

					case WindowMessage.CHANGECBCHAIN:
						if ( m.WParam == NextHandle )
						{
							ClearNextHandle();
							NextHandle = m.LParam;
						}
						else if ( NextHandle != IntPtr.Zero )
						{
							SendMessage( NextHandle, m.Msg, m.WParam, m.LParam );
						}
						break;
				}
				base.WndProc( ref m );
			}
		}

		#endregion
	}
}