﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;

namespace Tomham.Windows.Forms
{
	// Usage:
	// まずはSetSource
	// Sourceに変更があればUpdateMatchedItemsを呼んで表示の更新をする
	// フィルタ方法を変えたいならSetMatchedItemPredicaterを使う
	// Itemのソート方法はSetMatchedItemComparerで変更できる
	// Itemを得るにはGetMatchedItems等を使う

	public partial class FilterListBox : UserControl
	{
		public FilterListBox()
		{
			InitializeComponent();

			IncrementalSearch = true;
		}

		#region Properties

		private IMatchedItemPredicater MatchedItemPredicaterStore;
		private int MaxShowItemCountStore;

		[Browsable( false )]
		public string SelectedText
		{
			get { return FilterTextBox.SelectedText; }
			set { FilterTextBox.SelectedText = value; }
		}

		[Category( "Appearance" )]
		[Browsable( true )]
		public int MaxShowItemCount
		{
			get { return MaxShowItemCountStore; }
			set
			{
				if ( value < 0 )
				{
					throw new ArgumentOutOfRangeException();
				}

				MaxShowItemCountStore = value;
				UpdateMatchedItems();
			}
		}

		[Category( "Appearance" )]
		[Browsable( true )]
		public string FilterText
		{
			get { return FilterTextBox.Text; }
			set { FilterTextBox.Text = value; }
		}

		[Category( "Appearance" )]
		[Browsable( true )]
		public Color TextForeColor
		{
			get { return FilterTextBox.ForeColor; }
			set { SetTextForeColor( value ); }
		}

		[Category( "Appearance" )]
		[Browsable( true )]
		public Color TextBackColor
		{
			get { return FilterTextBox.BackColor; }
			set { SetTextBackColor( value ); }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public bool HorizontalScrollbar
		{
			get { return MatchedItemsListBox.HorizontalScrollbar; }
			set { MatchedItemsListBox.HorizontalScrollbar = value; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public int HorizontalExtent
		{
			get { return MatchedItemsListBox.HorizontalExtent; }
			set { MatchedItemsListBox.HorizontalExtent = value; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public SelectionMode SelectionMode
		{
			get { return MatchedItemsListBox.SelectionMode; }
			set { MatchedItemsListBox.SelectionMode = value; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public ImeMode TextBoxImeMode
		{
			get { return FilterTextBox.ImeMode; }
			set { FilterTextBox.ImeMode = value; }
		}

		[Browsable( false )]
		public bool TextBoxFocused
		{
			get { return FilterTextBox.Focused; }
		}

		[Browsable( false )]
		public bool ListBoxFoucused
		{
			get { return MatchedItemsListBox.Focused; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public ContextMenuStrip TextBoxContextMenuStrip
		{
			get { return FilterTextBox.ContextMenuStrip; }
			set { FilterTextBox.ContextMenuStrip = value; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		public ContextMenuStrip ListBoxContextMenuStrip
		{
			get { return MatchedItemsListBox.ContextMenuStrip; }
			set { MatchedItemsListBox.ContextMenuStrip = value; }
		}

		[Category( "Behavior" )]
		[Browsable( true )]
		[DefaultValue( true )]
		public bool IncrementalSearch { get; set; }

		public void SetMatchedItemPredicater( IMatchedItemPredicater value )
		{
			MatchedItemPredicaterStore = value;
			UpdateMatchedItems();
		}

		public IMatchedItemPredicater GetMatchedItemPredicater()
		{
			if ( MatchedItemPredicaterStore == null )
			{
				MatchedItemPredicaterStore = new DefaultMatchedItemPredicater();
			}

			return MatchedItemPredicaterStore;
		}

		private void SetTextForeColor( Color value )
		{
			FilterTextBox.ForeColor = value;
			MatchedItemsListBox.ForeColor = value;
		}

		private void SetTextBackColor( Color value )
		{
			FilterTextBox.BackColor = value;
			MatchedItemsListBox.BackColor = value;
		}

		#region Source

		private IEnumerable SourceStore;

		public IEnumerable GetSource()
		{
			if ( SourceStore == null )
			{
				SourceStore = new object[0];
			}
			return SourceStore;
		}

		public void SetSource( IEnumerable value )
		{
			if ( value == null )
			{
				SourceStore = new object[0];
			}
			else
			{
				SourceStore = value;
			}

			UpdateMatchedItems();
		}

		#endregion

		#endregion Properties

		#region MatchedItemComparer

		private IComparer MatchedItemComparerStore;

		public IComparer GetMatchedItemComparer()
		{
			if ( MatchedItemComparerStore == null )
			{
				return DefaultFilterListBoxItemComparer.Instance;
			}

			return MatchedItemComparerStore;
		}

		public void SetMatchedItemComparer( IComparer comparer )
		{
			MatchedItemComparerStore = comparer;

			UpdateMatchedItems();
		}

		#endregion

		#region ListBox Click Events

		[Category( "Action" )]
		[Browsable( true )]
		public event EventHandler ListBoxClick;

		protected virtual void OnListBoxClick( EventArgs e )
		{
			if ( ListBoxClick != null )
			{
				ListBoxClick( this, e );
			}
		}

		[Category( "Action" )]
		[Browsable( true )]
		public event EventHandler ListBoxDoubleClick;

		protected virtual void OnListBoxDoubleClick( EventArgs e )
		{
			if ( ListBoxDoubleClick != null )
			{
				ListBoxDoubleClick( this, e );
			}
		}

		[Category( "Action" )]
		[Browsable( true )]
		public event MouseEventHandler ListBoxMouseClick;

		protected virtual void OnListBoxMouseClick( MouseEventArgs e )
		{
			if ( ListBoxMouseClick != null )
			{
				ListBoxMouseClick( this, e );
			}
		}

		[Category( "Action" )]
		[Browsable( true )]
		public event MouseEventHandler ListBoxMouseDoubleClick;

		protected virtual void OnListBoxMouseDoubleClick( MouseEventArgs e )
		{
			if ( ListBoxMouseDoubleClick != null )
			{
				ListBoxMouseDoubleClick( this, e );
			}
		}

		private void MatchedItemsListBox_Click( object sender, EventArgs e )
		{
			OnListBoxClick( e );
		}

		private void MatchedItemsListBox_DoubleClick( object sender, EventArgs e )
		{
			OnListBoxDoubleClick( e );
		}

		private void MatchedItemsListBox_MouseClick( object sender, MouseEventArgs e )
		{
			OnListBoxMouseClick( e );
		}

		private void MatchedItemsListBox_MouseDoubleClick( object sender, MouseEventArgs e )
		{
			OnListBoxMouseDoubleClick( e );
		}

		#endregion ListBox Click Events

		#region FilterTextChanged Event

		private void FilterTextBox_TextChanged( object sender, EventArgs e )
		{
			if ( IncrementalSearch )
			{
				UpdateMatchedItems();
			}

			OnFilterTextChanged( e );
		}

		[Category( "Default" )]
		[Browsable( true )]
		public event EventHandler FilterTextChanged;

		protected virtual void OnFilterTextChanged( EventArgs e )
		{
			if ( FilterTextChanged != null )
			{
				FilterTextChanged( this, e );
			}
		}

		#endregion

		#region Key Events

		#region  Key Events with FilterTextBox

		private const Keys SelectAllKey = Keys.Control | Keys.A;

		[Category( "Key" )]
		[Browsable( true )]
		public event KeyPressEventHandler TextBoxKeyPress;

		protected virtual void OnTextBoxKeyPress( KeyPressEventArgs e )
		{
			if ( TextBoxKeyPress != null )
			{
				TextBoxKeyPress( this, e );
			}
		}

		[Category( "Key" )]
		[Browsable( true )]
		public event KeyEventHandler TextBoxKeyDown;

		protected virtual void OnTextBoxKeyDown( KeyEventArgs e )
		{
			if ( TextBoxKeyDown != null )
			{
				TextBoxKeyDown( this, e );
			}
		}

		[Category( "Key" )]
		[Browsable( true )]
		public event KeyEventHandler TextBoxKeyUp;

		protected virtual void OnTextBoxKeyUp( KeyEventArgs e )
		{
			if ( TextBoxKeyUp != null )
			{
				TextBoxKeyUp( this, e );
			}
		}

		private void FilterTextBox_KeyDown( object sender, KeyEventArgs e )
		{
			if ( ( e.KeyData ==
			       Keys.Enter ) &&
			     ( !IncrementalSearch ) )
			{
				UpdateMatchedItems();
				e.SuppressKeyPress = true;
			}
			else
			{
				switch ( e.KeyData )
				{
					case Keys.Down:
						ClearSelected();
						FocusToListBox();
						SelectFirstItem();
						e.SuppressKeyPress = true;
						break;

					case SelectAllKey:
						FilterTextBox.SelectAll();
						e.SuppressKeyPress = true;
						break;

					default:
						OnTextBoxKeyDown( e );
						break;
				}
			}
		}

		private void FilterTextBox_KeyPress( object sender, KeyPressEventArgs e )
		{
			OnTextBoxKeyPress( e );
		}

		private void FilterTextBox_KeyUp( object sender, KeyEventArgs e )
		{
			OnTextBoxKeyUp( e );
		}

		#endregion

		#region Key Events with MatchedItemsListBox

		[Category( "Key" )]
		[Browsable( true )]
		public event KeyPressEventHandler ListBoxKeyPress;

		protected virtual void OnListBoxKeyPress( KeyPressEventArgs e )
		{
			if ( ListBoxKeyPress != null )
			{
				ListBoxKeyPress( this, e );
			}
		}

		[Category( "Key" )]
		[Browsable( true )]
		public event KeyEventHandler ListBoxKeyDown;

		protected virtual void OnListBoxKeyDown( KeyEventArgs e )
		{
			if ( ListBoxKeyDown != null )
			{
				ListBoxKeyDown( this, e );
			}
		}

		[Category( "Key" )]
		[Browsable( true )]
		public event KeyEventHandler ListBoxKeyUp;

		protected virtual void OnListBoxKeyUp( KeyEventArgs e )
		{
			if ( ListBoxKeyUp != null )
			{
				ListBoxKeyUp( this, e );
			}
		}

		private void MatchedItemsListBox_KeyDown( object sender, KeyEventArgs e )
		{
			if ( e.KeyData ==
			     Keys.Up )
			{
				if ( HasSelectedFirstItemOrNoMatchedItems() )
				{
					e.SuppressKeyPress = true;
					FocusToTextBox();
				}
			}
			else
			{
				OnListBoxKeyDown( e );
			}
		}

		private bool HasSelectedFirstItemOrNoMatchedItems()
		{
			if ( !HasMatchedItems() )
			{
				return true;
			}
			if ( GetSelectedItemCount() == 1 )
			{
				return MatchedItemsListBox.SelectedIndex == 0;
			}

			return false;
		}

		private void MatchedItemsListBox_KeyPress( object sender, KeyPressEventArgs e )
		{
			OnListBoxKeyPress( e );
		}

		private void MatchedItemsListBox_KeyUp( object sender, KeyEventArgs e )
		{
			OnListBoxKeyUp( e );
		}

		#endregion

		#endregion Key Events

		#region Enter and Leave Events

		[Category( "Focus" )]
		[Browsable( true )]
		public event EventHandler TextBoxEnter;

		protected virtual void OnTextBoxEnter( EventArgs e )
		{
			if ( TextBoxEnter != null )
			{
				TextBoxEnter( this, e );
			}
		}

		[Category( "Focus" )]
		[Browsable( true )]
		public event EventHandler TextBoxLeave;

		protected virtual void OnTextBoxLeave( EventArgs e )
		{
			if ( TextBoxLeave != null )
			{
				TextBoxLeave( this, e );
			}
		}

		[Category( "Focus" )]
		[Browsable( true )]
		public event EventHandler ListBoxEnter;

		protected virtual void OnListBoxEnter( EventArgs e )
		{
			if ( ListBoxEnter != null )
			{
				ListBoxEnter( this, e );
			}
		}

		[Category( "Focus" )]
		[Browsable( true )]
		public event EventHandler ListBoxLeave;

		protected virtual void OnListBoxLeave( EventArgs e )
		{
			if ( ListBoxLeave != null )
			{
				ListBoxLeave( this, e );
			}
		}

		private void FilterTextBox_Enter( object sender, EventArgs e )
		{
			OnTextBoxEnter( e );
		}

		private void FilterTextBox_Leave( object sender, EventArgs e )
		{
			OnTextBoxLeave( e );
		}

		private void MatchedItemsListBox_Enter( object sender, EventArgs e )
		{
			OnListBoxEnter( e );
		}

		private void MatchedItemsListBox_Leave( object sender, EventArgs e )
		{
			OnListBoxLeave( e );
		}

		#endregion

		[Category( "Behavior" )]
		[Browsable( true )]
		public event EventHandler SelectedIndexChanged;

		protected virtual void OnSelectedIndexChanged( EventArgs e )
		{
			if ( SelectedIndexChanged != null )
			{
				SelectedIndexChanged( this, e );
			}
		}

		private void MatchedItemsListBox_SelectedIndexChanged( object sender, EventArgs e )
		{
			OnSelectedIndexChanged( e );
		}

		public bool HasMatchedItems()
		{
			return 0 < MatchedItemsListBox.Items.Count;
		}

		public int GetMatchedItemCount()
		{
			return MatchedItemsListBox.Items.Count;
		}

		public int GetSelectedItemCount()
		{
			return MatchedItemsListBox.SelectedItems.Count;
		}

		public void SetSelected( int index, bool value )
		{
			MatchedItemsListBox.SetSelected( index, value );
		}

		public void SelectAllItems()
		{
			MatchedItemsListBox.SelectAll();
		}

		public void SelectFirstItem()
		{
			if ( HasMatchedItems() )
			{
				SetSelected( 0, true );
			}
		}

		public bool HasSelectedItems()
		{
			return 0 < MatchedItemsListBox.SelectedItems.Count;
		}

		public void ClearSelected()
		{
			MatchedItemsListBox.ClearSelected();
		}

		public bool FocusToTextBox()
		{
			return FilterTextBox.Focus();
		}

		public bool FocusToListBox()
		{
			return MatchedItemsListBox.Focus();
		}

		public void ClearFilterText()
		{
			FilterTextBox.Text = string.Empty;
		}

		#region GetItems Methods

		public object GetFirstItem()
		{
			if ( HasMatchedItems() )
			{
				return MatchedItemsListBox.Items[0];
			}

			return null;
		}

		public object GetSelectedItem()
		{
			return MatchedItemsListBox.SelectedItem;
		}

		public object GetOnlyOneSelectedItem()
		{
			if ( GetSelectedItemCount() == 1 )
			{
				return MatchedItemsListBox.SelectedItem;
			}

			return null;
		}

		public List<T> GetSelectedItems<T>()
		{
			return MatchedItemsListBox.SelectedItems.Cast<T>().ToList();
		}

		public List<T> GetMatchedItems<T>()
		{
			return MatchedItemsListBox.Items.Cast<T>().ToList();
		}

		#endregion

		#region UpdateMatchedItems

		[Browsable( true )]
		public event EventHandler<EventArgs> MatchedItemsUpdating;

		protected virtual void OnMatchedItemsUpdating( EventArgs e )
		{
			if ( MatchedItemsUpdating != null )
			{
				MatchedItemsUpdating( this, e );
			}
		}

		[Browsable( true )]
		public event EventHandler<EventArgs> MatchedItemsUpdated;

		protected virtual void OnMatchedItemsUpdated( EventArgs e )
		{
			if ( MatchedItemsUpdated != null )
			{
				MatchedItemsUpdated( this, e );
			}
		}

		public void UpdateMatchedItems()
		{
			OnMatchedItemsUpdating( EventArgs.Empty );

			DoUpdateMatchedItems();

			OnMatchedItemsUpdated( EventArgs.Empty );
		}

		private void DoUpdateMatchedItems()
		{
			object[] matchedItems = FindMatchedItems();

			Array.Sort( matchedItems, GetMatchedItemComparer() );

			UpdateMatchedItemsListBox( matchedItems );
		}

		private void UpdateMatchedItemsListBox( object[] matchedItems )
		{
			MatchedItemsListBox.BeginUpdate();

			MatchedItemsListBox.Items.Clear();

			AddItemsToMatchedItemsListBox( matchedItems );

			SelectFirstItem();

			MatchedItemsListBox.EndUpdate();
		}

		private void AddItemsToMatchedItemsListBox( object[] matchedItems )
		{
			if ( 0 < MaxShowItemCount )
			{
				IEnumerable<object> items = matchedItems.Take( MaxShowItemCount );

				MatchedItemsListBox.Items.AddRange( items.ToArray() );
			}
			else
			{
				MatchedItemsListBox.Items.AddRange( matchedItems );
			}
		}

		#region FindMatchedItems

		private object[] FindMatchedItems()
		{
			object[] source = EnumerableToArray( GetSource() );

			var result = new List<object>();

			foreach ( object item in source )
			{
				if ( GetMatchedItemPredicater().IsMatchedItem( FilterText, item ) )
				{
					result.Add( item );
				}
			}

			return result.ToArray();
		}

		private object[] EnumerableToArray( IEnumerable enumerable )
		{
			var list = new List<object>();

			foreach ( object item in enumerable )
			{
				list.Add( item );
			}

			return list.ToArray();
		}

		#endregion

		#endregion
	}
}