﻿using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Tomham.Windows.Forms
{
	public class HotKey : IDisposable
	{
		private readonly HotKeyForm Dummy;
		private readonly int ID;
		private CombinationKeys CombinationKeysStore;

		public HotKey()
		{
			Dummy = new HotKeyForm( this );
			ID = Dummy.GetType().GetHashCode();
			Enabled = true;
		}

		public bool Enabled { get; set; }

		public object Tag { get; set; }

		public CombinationKeys CombinationKeys
		{
			get { return CombinationKeysStore; }
			set
			{
				if ( CombinationKeys.Equals( value ) )
				{
					if ( !HasRegistered &&
					     IsValid( CombinationKeys ) )
					{
						RegisterHotKey();
					}

					return;
				}

				if ( HasRegistered )
				{
					UnRegisterHotKey();
				}

				SetCombinationKeys( value );

				if ( IsValid( CombinationKeys ) )
				{
					RegisterHotKey();
				}
			}
		}

		public bool HasRegistered { get; private set; }

		#region IDisposable Members

		public void Dispose()
		{
			if ( HasRegistered )
			{
				UnRegisterHotKey();
			}
			Dummy.Dispose();
		}

		#endregion

		private void SetCombinationKeys( CombinationKeys value )
		{
			if ( IsValid( value ) )
			{
				CombinationKeysStore = value;
			}
			else
			{
				ClearCombinationKeysStore();
			}
		}

		private bool IsValid( CombinationKeys keys )
		{
			return keys.KeyCode != Keys.None;
		}

		private void ClearCombinationKeysStore()
		{
			CombinationKeysStore = new CombinationKeys();
		}

		private void RegisterHotKey()
		{
			HasRegistered = RegisterHotKey(
				Dummy.Handle,
				ID,
				(int) CombinationKeys.Modifiers,
				(int) CombinationKeys.KeyCode
				);
		}

		private void UnRegisterHotKey()
		{
			HasRegistered = false;
			UnregisterHotKey( Dummy.Handle, ID );
		}

		public event EventHandler Pressed;

		protected virtual void OnPressed( EventArgs e )
		{
			if ( Pressed != null )
			{
				Pressed( this, e );
			}
		}

		[DllImport( "user32.dll" )]
		private static extern bool RegisterHotKey( IntPtr hWnd, int id, int modifiers, int virtualKey );

		[DllImport( "user32.dll" )]
		private static extern bool UnregisterHotKey( IntPtr hWnd, int id );

		#region Nested type: HotKeyForm

		private class HotKeyForm : Form
		{
			private readonly HotKey HotKey;

			public HotKeyForm( HotKey hotKey )
			{
				HotKey = hotKey;
			}

			protected override void WndProc( ref Message m )
			{
				switch ( m.Msg )
				{
					case WindowMessage.HOTKEY:
						if ( m.WParam ==
						     (IntPtr) HotKey.ID )
						{
							if ( HotKey.Enabled )
							{
								HotKey.OnPressed( EventArgs.Empty );
							}
						}

						break;
				}
				base.WndProc( ref m );
			}
		}

		#endregion
	}
}