﻿using System.Collections.Generic;
using System.Windows.Forms;

namespace Tomham.Windows.Forms
{
	public static class KeysHelper
	{
		public static bool IsModifiersOnly( Keys keys )
		{
			if ( ExtractKeyCode( keys ) !=
			     Keys.None )
			{
				return false;
			}

			return ContainsAlt( keys ) || ContainsControl( keys ) || ContainsShift( keys );
		}

		public static bool IsModifiersOnly( CombinationKeys keys )
		{
			if ( keys.KeyCode !=
			     Keys.None )
			{
				return false;
			}

			return keys.Modifiers != Modifiers.None;
		}

		public static bool ContainsShift( Keys keys )
		{
			return Contains( keys, Keys.Shift );
		}

		public static bool ContainsControl( Keys keys )
		{
			return Contains( keys, Keys.Control );
		}

		public static bool ContainsAlt( Keys keys )
		{
			return Contains( keys, Keys.Alt );
		}

		public static bool Contains( Keys keys, Keys modifiers )
		{
			return ( modifiers & keys ) == modifiers;
		}

		public static Keys ExtractKeyCode( Keys keys )
		{
			return keys & Keys.KeyCode;
		}

		public static bool ContainsAlt( Modifiers modifiers )
		{
			return Contains( modifiers, Modifiers.Alt );
		}

		public static bool ContainsControl( Modifiers modifiers )
		{
			return Contains( modifiers, Modifiers.Control );
		}

		public static bool ContainsShift( Modifiers modifiers )
		{
			return Contains( modifiers, Modifiers.Shift );
		}

		public static bool ContainsWindows( Modifiers modifiers )
		{
			return Contains( modifiers, Modifiers.Windows );
		}

		/// <summary>
		/// x contains y?
		/// </summary>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <returns></returns>
		public static bool Contains( Modifiers x, Modifiers y )
		{
			return ( x & y ) == y;
		}

		public static CombinationKeys Convert( Keys keys )
		{
			Modifiers modifiers = Modifiers.None;
			if ( ContainsAlt( keys ) )
			{
				modifiers |= Modifiers.Alt;
			}
			if ( ContainsControl( keys ) )
			{
				modifiers |= Modifiers.Control;
			}
			if ( ContainsShift( keys ) )
			{
				modifiers |= Modifiers.Shift;
			}

			return new CombinationKeys { KeyCode = ExtractKeyCode( keys ), Modifiers = modifiers };
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="keys"></param>
		/// <returns></returns>
		/// <remarks>WindowsKeyは反映されない</remarks>
		public static Keys Convert( CombinationKeys keys )
		{
			Keys result = keys.KeyCode;
			if ( ContainsAlt( keys.Modifiers ) )
			{
				result |= Keys.Alt;
			}
			if ( ContainsControl( keys.Modifiers ) )
			{
				result |= Keys.Control;
			}
			if ( ContainsShift( keys.Modifiers ) )
			{
				result |= Keys.Shift;
			}

			return result;
		}

		#region GetKeyCodes

		public static List<Keys> GetKeyCodes()
		{
			var result = new List<Keys>();

			AddAlphabets( result );

			AddNumbers( result );

			AddFunctionKeys( result );

			AddEtc( result );

			AddNumPadKeys( result );

			return result;
		}

		private static void AddAlphabets( List<Keys> to )
		{
			AddItems( to, 65, 90 );
		}

		private static void AddNumbers( List<Keys> to )
		{
			AddItems( to, 48, 57 );
		}

		private static void AddFunctionKeys( List<Keys> to )
		{
			AddItems( to, 0x70, 0x87 );
		}

		private static void AddNumPadKeys( List<Keys> to )
		{
			AddItems( to, 0x60, 0x6F );
		}

		private static void AddEtc( List<Keys> to )
		{
			to.Add( Keys.Back );
			to.Add( Keys.Tab );
			to.Add( Keys.Enter );
			to.Add( Keys.Pause );
			to.Add( Keys.Escape );

			to.Add( Keys.IMEConvert );
			to.Add( Keys.IMENonconvert );

			AddItems( to, 0x20, 0x28 );
			AddItems( to, 0x2C, 0x2E );

			AddItems( to, 0xba, 0xBF );

			// @ key
			to.Add( (Keys) 0xC0 );

			AddItems( to, 0xDB, 0xDE );
			to.Add( Keys.Oem102 );

			AddItems( to, 0x90, 0x91 );
		}

		private static void AddItems( List<Keys> to, int startCode, int endCode )
		{
			for ( int i = startCode; i <= endCode; i++ )
			{
				to.Add( (Keys) i );
			}
		}

		#endregion
	}
}