﻿using System.Collections.Generic;
using System.Windows.Forms;

namespace Tomham.Windows.Forms
{
	/// <summary>
	/// 
	/// </summary>
	/// <remarks>ListBox.Sortedの事は全く考えてません</remarks>
	public static class ListBoxHelper
	{
		public static void MoveUpSelectedItems( this ListBox listBox )
		{
			int[] selectedIndices = GetSelectedIndicesCopy( listBox );

			listBox.BeginUpdate();

			foreach ( int index in selectedIndices )
			{
				MoveUp( index, listBox );
			}

			listBox.EndUpdate();
		}

		private static void MoveUp( int fromIndex, ListBox listBox )
		{
			int toIndex = fromIndex - 1;

			if ( HasToNeedToMove( toIndex, listBox ) )
			{
				SwapItems( listBox, fromIndex, toIndex );
			}
		}

		public static void SwapItems( this ListBox listBox, int index1, int index2 )
		{
			listBox.BeginUpdate();

			bool oldIndex1Selected = listBox.GetSelected( index1 );
			bool oldIndex2Selected = listBox.GetSelected( index2 );

			DoSwapItems( listBox, index1, index2 );

			listBox.SetSelected( index1, oldIndex2Selected );
			listBox.SetSelected( index2, oldIndex1Selected );

			listBox.EndUpdate();
		}

		private static void DoSwapItems( ListBox listBox, int index1, int index2 )
		{
			object object1 = listBox.Items[index1];
			object object2 = listBox.Items[index2];

			listBox.Items[index2] = object1;
			listBox.Items[index1] = object2;
		}

		private static bool HasToNeedToMove( int toIndex, ListBox listBox )
		{
			if ( !IsValidIndex( toIndex, listBox ) )
			{
				return false;
			}

			if ( listBox.GetSelected( toIndex ) )
			{
				return false;
			}

			return true;
		}

		private static bool IsValidIndex( int index, ListBox listBox )
		{
			if ( index < 0 )
			{
				return false;
			}

			if ( listBox.Items.Count <= index )
			{
				return false;
			}

			return true;
		}

		private static int[] GetSelectedIndicesCopy( ListBox listBox )
		{
			var selectedIndices = new int[listBox.SelectedItems.Count];

			listBox.SelectedIndices.CopyTo( selectedIndices, 0 );

			return selectedIndices;
		}

		public static void MoveDownSelectedItems( this ListBox listBox )
		{
			int[] selectedIndices = GetSelectedIndicesCopy( listBox );

			listBox.BeginUpdate();

			for ( int i = selectedIndices.Length - 1; 0 <= i; i-- )
			{
				int index = selectedIndices[i];

				MoveDown( index, listBox );
			}

			listBox.EndUpdate();
		}

		private static void MoveDown( int fromIndex, ListBox listBox )
		{
			int toIndex = fromIndex + 1;

			if ( HasToNeedToMove( toIndex, listBox ) )
			{
				SwapItems( listBox, fromIndex, toIndex );
			}
		}

		public static void RemoveSelectedItems( this ListBox listBox )
		{
			listBox.BeginUpdate();

			while ( 0 <= listBox.SelectedIndex )
			{
				listBox.Items.RemoveAt( listBox.SelectedIndex );
			}

			listBox.EndUpdate();
		}

		public static void MoveSelectedItemsToOther( this ListBox listBox, ListBox other )
		{
			if ( !listBox.HasSelectedItems() )
			{
				return;
			}

			List<object> selectedItems = GetSelectedItems( listBox );

			listBox.RemoveSelectedItems();

			other.BeginUpdate();

			if ( other.HasSelectedItems() )
			{
				int firstIndex = GetFirstIndex( other );

				IncertToOther( other, firstIndex, selectedItems );

				SelectMovedItems( other, firstIndex, selectedItems.Count );
			}
			else
			{
				int index = other.Items.Count;
				other.Items.AddRange( selectedItems.ToArray() );
				SelectMovedItems( other, index, selectedItems.Count );
			}

			other.EndUpdate();
		}

		private static List<object> GetSelectedItems( ListBox listBox )
		{
			var result = new List<object>();
			foreach ( object selectedItem in listBox.SelectedItems )
			{
				result.Add( selectedItem );
			}
			return result;
		}

		private static void IncertToOther( ListBox other, int index, List<object> selectedItems )
		{
			for ( int i = selectedItems.Count - 1; i >= 0; i-- )
			{
				other.Items.Insert( index, selectedItems[i] );
			}
		}

		private static void SelectMovedItems( ListBox other, int index, int count )
		{
			var indices = new List<int>();

			for ( int i = 0; i < count; i++ )
			{
				indices.Add( index );
				index++;
			}

			other.SelectedItems.Clear();

			foreach ( int selectedIndex in indices )
			{
				other.SetSelected( selectedIndex, true );
			}
		}

		private static int GetFirstIndex( ListBox other )
		{
			int index;

			if ( other.HasSelectedItems() )
			{
				index = other.SelectedIndex;
			}
			else
			{
				index = 0;
			}
			return index;
		}

		public static bool HasSelectedItems( this ListBox listBox )
		{
			return 0 < listBox.SelectedItems.Count;
		}

		public static void ReplaceItem( this ListBox listBox, int index, object newItem )
		{
			bool selected = listBox.SelectedIndices.Contains( index );

			listBox.BeginUpdate();
			listBox.Items.RemoveAt( index );
			listBox.Items.Insert( index, newItem );
			listBox.SetSelected( index, selected );
			listBox.EndUpdate();
		}

		public static void ReplaceSelectedItem( this ListBox listBox, object newItem )
		{
			ReplaceItem( listBox, listBox.SelectedIndex, newItem );
		}

		public static void ClearAndAddRange( ListBox to, object[] addition )
		{
			to.Items.Clear();
			to.Items.AddRange( addition );
		}

		public static void SelectAll( this ListBox listBox )
		{
			for ( int index = 0; index < listBox.Items.Count; index++ )
			{
				listBox.SetSelected( index, true );
			}
		}

		#region Reverse

		public static void Reverse( this ListBox listBox )
		{
			List<object> reversedItems = GetReversedItems( listBox );

			List<bool> reversedSelectedStates = GetReversedSelectedStates( listBox );

			listBox.BeginUpdate();

			ClearAndAddRange( listBox, reversedItems.ToArray() );

			SetSelectedStates( listBox, reversedSelectedStates );

			listBox.EndUpdate();
		}

		private static void SetSelectedStates( ListBox listBox, List<bool> selectedStates )
		{
			for ( int index = 0; index < selectedStates.Count; index++ )
			{
				listBox.SetSelected( index, selectedStates[index] );
			}
		}

		private static List<bool> GetReversedSelectedStates( ListBox listBox )
		{
			var selectedStates = new List<bool>();

			for ( int index = 0; index < listBox.Items.Count; index++ )
			{
				selectedStates.Add( listBox.GetSelected( index ) );
			}

			selectedStates.Reverse();
			return selectedStates;
		}

		private static List<object> GetReversedItems( ListBox listBox )
		{
			var items = new List<object>();
			foreach ( object item in listBox.Items )
			{
				items.Add( item );
			}

			items.Reverse();
			return items;
		}

		#endregion
	}
}